/*
 * STRUMPACK -- STRUctured Matrices PACKage, Copyright (c) 2014, The Regents of
 * the University of California, through Lawrence Berkeley National Laboratory
 * (subject to receipt of any required approvals from the U.S. Dept. of Energy).
 * All rights reserved.
 *
 * If you have questions about your rights to use or distribute this software,
 * please contact Berkeley Lab's Technology Transfer Department at TTD@lbl.gov.
 *
 * NOTICE. This software is owned by the U.S. Department of Energy. As such, the
 * U.S. Government has been granted for itself and others acting on its behalf a
 * paid-up, nonexclusive, irrevocable, worldwide license in the Software to
 * reproduce, prepare derivative works, and perform publicly and display publicly.
 * Beginning five (5) years after the date permission to assert copyright is
 * obtained from the U.S. Department of Energy, and subject to any subsequent five
 * (5) year renewals, the U.S. Government is granted for itself and others acting
 * on its behalf a paid-up, nonexclusive, irrevocable, worldwide license in the
 * Software to reproduce, prepare derivative works, distribute copies to the
 * public, perform publicly and display publicly, and to permit others to do so.
 *
 * Developers: Francois-Henry Rouet, Xiaoye S. Li, Pieter Ghysels
 *             (Lawrence Berkeley National Lab, Computational Research Division).
 *
 */

#ifndef HSS_PAR_PRODUCT_HPP
#define HSS_PAR_PRODUCT_HPP

template<typename T, typename S> void HSS_tree<T,S>::HSS_par_Product(char Trans, T alpha, T *X, int *descX, int d, T beta, T *B, int * descB, gridinfo *ginfo, int partial) {
  /* HSS matrix-vector product B= alpha*op(A)*X+beta*B with a vector X with d columns.
   * op(A)=A is Trans='N', op(A)=A* is Trans='C'.
   * If IdGen=true, generators U and V are assumed to be identity (this is useful for
   * the adaptive rank mechanism).
   *
   * Algorithm:
   * Step1: postorder traversal:
   *   if i leaf, y{i}=V{i}'*x(I,:)
   *   otherwise, y{i}=V{i}'*[y{child1};y{child2}]
   * Step2: reverse postorder traversal
   *   if i leaf, b(I,:)=U{i}*z{i}+D{i}*x(I,:)
   *   otherwise,
   *     r1/2: row-rank for child1/2
   *     z{child1}=B12{i}*y{child2}+U{i}(1:r1)*z{i}
   *     z{child2}=B21{i}*y{child1}+U{i}(r1+1:r1+r2)*z{i}
   *     (if root node, the second term in the sum is not here).
   *
   * If "partial": traverse only the rightmost subtree. This is
   * useful for computing products with the Schur complement.
   *
   * The implementation is not very smart in terms of data structure; each node
   * has a "y" and a "z" as in the algorithm above.
   *
   * The parallel version follows the same flow as the parallel compression;
   * there are three kinds of nodes to distinguish: serial subtrees, leaves
   * that are mapped on several processes, and parallel interior nodes.
   *
   */

  int i, j, k;
  int mb, nb;
  int ib;
  int br, b;
  int rows_y, rows_z, rows_y1, rows_y2, rows_z1, rows_z2;
  int rows_B, cols_B;
  int istart, iend;
  int first, last;
  int ldl;
  T *tmp, *tmp2, *Xl, *Bl;
  T **y, **z;
  int *children1, *children2;
  int ch1, ch2;
  int myid, np;
  int proc_b, proc_g;
  int idummy, ierr;
  int *descY=NULL, *descZ=NULL, *descBX=NULL, descGen[BLACSCTXTSIZE], *desc=NULL;
  int locr, locc;
  HSS_node<T,S> *child1, *child2;
  gridinfo tmpinfo, tmpinfo1, tmpinfo2;

  mb=ginfo->mb;
  nb=ginfo->nb;
  MPI_Comm_rank(comm,&myid);
  MPI_Comm_size(comm,&np);

  /* Here y and z are not fields of a node but arrays indexed following the
   * postorder; the pointers "children" of a node are not enough to find the
   * right in y and z, what we need is numbers in the postorder. "children1"
   * and "children2" are the same information as children but they provide
   * numbers (indices in the tree).
   *
   * children1: children[i]==j: node number j is left child of node number i.
   * child1: pointer to left child (type HSS_node) of node nodes[i].
   */

  children1=new int[dim_tree];
  for(i=0;i<dim_tree;i++) {
    children1[i]=-1;
    if(nodes[i]->nr_children!=0)
      for(j=0;j<dim_tree;j++)
        if(nodes[j]==nodes[i]->children[0]) {
          children1[i]=j;
          break;
        }
  }
  children2=new int[dim_tree];
  for(i=0;i<dim_tree;i++) {
    children2[i]=-1;
    if(nodes[i]->nr_children!=0)
      for(j=0;j<dim_tree;j++)
        if(nodes[j]==nodes[i]->children[1]) {
          children2[i]=j;
          break;
        }
  }

  /* As in the compression routine, the first thing to do is to redistribute
   * the input following the clustering induced by the parallel leaves of
   * the tree.
   */
  distVectorTree(X,descX,d,&Xl,&descBX,ginfo,nodes[dim_tree-1]->allctxt);
  delete[] descBX;
  distVectorTree(B,descB,d,&Bl,&descBX,ginfo,nodes[dim_tree-1]->allctxt);

  if(partial) {
    /* Traverse only the subtree rooted at the right child
     * of the root node. Its first node is the node
     * that follows the left child of the root node
     * in the postorder.
     */
    first=knode+1;
    last=dim_tree-2;
  } else {
    /* Traverse the whole tree. */
    first=0;
    last=dim_tree-1;
  }

  /* Step 1: postorder traversal (except root) */
  y=new T*[dim_tree-1];
  for(i=0;i<dim_tree-1;i++)
    y[i]=NULL;

  /* Serial subtree */
  if(subtree[2*myid]>=first) {
    istart=subtree[2*myid];
    iend=std::min(subtree[2*myid+1],dim_tree-2);
    ldl=np==1?a_dim:nodes[iend]->tree_ie-nodes[iend]->tree_ib+1;
    for(i=istart;i<=iend;i++) {
      /* Traverse subtree, excluding root node */
      ib=nodes[i]->tree_ib-nodes[istart]->tree_ib;
      rows_y=Trans=='N'?nodes[i]->v_rank:nodes[i]->u_rank;
      y[i]=new T[rows_y*d]();
      if(nodes[i]->nr_children==0) {
        /* Leaf: y{i}=V{i}'*x(I,:) */
        nodes[i]->seq_applyGenerator('C',Trans=='N'?'V':'U',d,alpha,&Xl[ib],ldl,ZERO,y[i],rows_y);
      } else {
        /* Non-leaf: y{i}=V{i}'*[y{child1};y{child2}] */
        child1=nodes[i]->children[0];
        child2=nodes[i]->children[1];
        ch1=children1[i];
        ch2=children2[i];

        rows_y1=Trans=='N'?child1->v_rank:child1->u_rank;
        rows_y2=Trans=='N'?child2->v_rank:child2->u_rank;
        tmp=new T[(rows_y1+rows_y2)*d]();
        lacpy('N',rows_y1,d,y[ch1],rows_y1,tmp,rows_y1+rows_y2);
        lacpy('N',rows_y2,d,y[ch2],rows_y2,&tmp[rows_y1],rows_y1+rows_y2);
        nodes[i]->seq_applyGenerator('C',Trans=='N'?'V':'U',d,ONE,tmp,rows_y1+rows_y2,ZERO,y[i],rows_y);
        delete[] tmp;
      }
    }
    if(np>1) {
      rows_y=Trans=='N'?nodes[iend]->v_rank:nodes[iend]->u_rank;
      idummy=std::max(1,numroc_(&rows_y,&mb,&nodes[iend]->myrow,&IZERO,&nodes[iend]->nprow));
      descY=new int[BLACSCTXTSIZE];
      descinit_(descY,&rows_y,&d,&mb,&nb,&IZERO,&IZERO,&(nodes[iend]->context),&idummy,&ierr);
    }
  }

  /* Leaves (avoid the case leaf=root) */
  for(i=first;i<last;i++) {
    if(   nodes[i]->procs[1]>nodes[i]->procs[0]
       && nodes[i]->nr_children==0
       && nodes[i]->procs[0]+nodes[i]->nprow*nodes[i]->npcol-1>=myid
       && nodes[i]->procs[0]<=myid
      ) {
      tmpinfo.ctxt=nodes[i]->context;
      tmpinfo.nprow=nodes[i]->nprow;
      tmpinfo.npcol=nodes[i]->npcol;
      tmpinfo.myrow=nodes[i]->myrow;
      tmpinfo.mycol=nodes[i]->mycol;
      tmpinfo.mb=ginfo->mb;
      tmpinfo.nb=ginfo->nb;
      rows_y=Trans=='N'?nodes[i]->v_rank:nodes[i]->u_rank;
      locr=numroc_(&rows_y,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow);
      locc=numroc_(&d,&nb,&nodes[i]->mycol,&IZERO,&nodes[i]->npcol);
      y[i]=new T[locr*locc]();
      idummy=std::max(1,locr);
      descY=new int[BLACSCTXTSIZE];
      descinit_(descY,&rows_y,&d,&mb,&nb,&IZERO,&IZERO,&nodes[i]->context,&idummy,&ierr);
      nodes[i]->par_applyGenerator('C',Trans=='N'?'V':'U',d,alpha,Xl,descBX,ZERO,y[i],descY,&tmpinfo);
    }
  }

  /* Interior nodes */
  for(j=0;j<par_lvl;j++) {
    i=par_nodes[j];
    if(i<first || i>=last)
      break;
    child1=nodes[i]->children[0];
    child2=nodes[i]->children[1];
    ch1=children1[i];
    ch2=children2[i];
    proc_b=nodes[i]->procs[0];
    proc_g=nodes[i]->procs[0]+nodes[i]->nprow*nodes[i]->npcol-1;
    tmpinfo.ctxt=nodes[i]->context;
    tmpinfo.nprow=nodes[i]->nprow;
    tmpinfo.npcol=nodes[i]->npcol;
    tmpinfo.myrow=nodes[i]->myrow;
    tmpinfo.mycol=nodes[i]->mycol;
    tmpinfo.mb=ginfo->mb;
    tmpinfo.nb=ginfo->nb;
    /* Merge y{child1} and y{child2} into tmp2 */
    if(myid<child1->procs[0]+child1->nprow*child1->npcol)
      /* Mapped on child 1 */
      tmp=y[ch1];
    else if(myid<child2->procs[0]+child2->nprow*child2->npcol)
      /* Mapped on child 2 */
      tmp=y[ch2];
    else
      /* Out of both grids */
      tmp=NULL;
    rows_y1=Trans=='N'?child1->v_rank:child1->u_rank;
    rows_y2=Trans=='N'?child2->v_rank:child2->u_rank;
    mergeVector(tmp,child1->procs[0],child1->procs[0]+child1->nprow*child1->npcol-1,rows_y1,IONE,child2->procs[0],child2->procs[0]+child2->nprow*child2->npcol-1,rows_y2,IONE,proc_b,proc_g,d,descY,&tmp2,&desc,nodes[i]->allctxt,&tmpinfo,comm);

    /* Exclude processes that are not in the grid of i */
    if(myid>=proc_b && myid<=proc_g) {
      rows_y=Trans=='N'?nodes[i]->v_rank:nodes[i]->u_rank;
      locr=numroc_(&rows_y,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow);
      locc=numroc_(&d,&nb,&nodes[i]->mycol,&IZERO,&nodes[i]->npcol);
      y[i]=new T[locr*locc]();
      idummy=std::max(1,locr);
      if(descY==NULL)
        descY=new int[BLACSCTXTSIZE];
      descinit_(descY,&rows_y,&d,&mb,&nb,&IZERO,&IZERO,&nodes[i]->context,&idummy,&ierr);
      nodes[i]->par_applyGenerator('C',Trans=='N'?'V':'U',d,ONE,tmp2,desc,ZERO,y[i],descY,&tmpinfo);
      delete[] tmp2;
    }
  }

  /* Step 2: reverse postorder traversal */
  z=new T*[dim_tree-1];
  for(i=0;i<dim_tree-1;i++)
    z[i]=NULL;

  /* Interior nodes */
  for(j=par_lvl-1;j>=0;j--) {
    i=par_nodes[j];
    if(i>last)
      continue;
    if(i<first)
      break;
    child1=nodes[i]->children[0];
    child2=nodes[i]->children[1];
    ch1=children1[i];
    ch2=children2[i];
    proc_b=nodes[i]->procs[0];
    proc_g=nodes[i]->procs[0]+nodes[i]->nprow*nodes[i]->npcol-1;
    tmpinfo.ctxt=nodes[i]->context;
    tmpinfo.nprow=nodes[i]->nprow;
    tmpinfo.npcol=nodes[i]->npcol;
    tmpinfo.myrow=nodes[i]->myrow;
    tmpinfo.mycol=nodes[i]->mycol;
    tmpinfo.mb=ginfo->mb;
    tmpinfo.nb=ginfo->nb;
    rows_z1=Trans=='N'?child1->u_rank:child1->v_rank;
    rows_z2=Trans=='N'?child2->u_rank:child2->v_rank;
    rows_y1=Trans=='N'?child1->v_rank:child1->u_rank;
    rows_y2=Trans=='N'?child2->v_rank:child2->u_rank;
    br=rows_z1+rows_z2;

    /* Merge y{child1} and y{child2} into tmp2 */
    if(myid>=child1->procs[0] && myid<child1->procs[0]+child1->nprow*child1->npcol) {
      /* Mapped on child 1 */
      tmp=y[ch1];
      idummy=std::max(1,numroc_(&rows_y1,&mb,&child1->myrow,&IZERO,&child1->nprow));
      descinit_(descY,&rows_y1,&d,&mb,&nb,&IZERO,&IZERO,&child1->context,&idummy,&ierr);
    } else if(myid>=child2->procs[0] && myid<child2->procs[0]+child2->nprow*child2->npcol) {
      /* Mapped on child 2 */
      tmp=y[ch2];
      idummy=std::max(1,numroc_(&rows_y2,&mb,&child2->myrow,&IZERO,&child2->nprow));
      descinit_(descY,&rows_y2,&d,&mb,&nb,&IZERO,&IZERO,&child2->context,&idummy,&ierr);
    } else {
      /* Out of both grids */
      tmp=NULL;
      descY=NULL;
    }

    mergeVector(tmp,child1->procs[0],child1->procs[0]+child1->nprow*child1->npcol-1,rows_y1,IONE,child2->procs[0],child2->procs[0]+child2->nprow*child2->npcol-1,rows_y2,IONE,proc_b,proc_g,d,descY,&tmp2,&desc,nodes[i]->allctxt,&tmpinfo,comm);

    delete[] descY;
    descY=desc;

    /* Exclude processes that are not in the grid of i */
    if(myid>=proc_b && myid<=proc_g) {
      locr=numroc_(&br,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow);
      locc=numroc_(&d,&nb,&nodes[i]->mycol,&IZERO,&nodes[i]->npcol);
      tmp=new T[locr*locc]();
      idummy=std::max(1,locr);
      descZ=new int[BLACSCTXTSIZE];
      descinit_(descZ,&br,&d,&mb,&nb,&IZERO,&IZERO,&nodes[i]->context,&idummy,&ierr);

      /* Child 1: z{child1}=B12{i}*y{child2} */
      rows_B=Trans=='N'?child1->u_rank:child2->u_rank;
      cols_B=Trans=='N'?child2->v_rank:child1->v_rank;
      idummy=std::max(1,numroc_(&rows_B,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow));
      descinit_(descGen,&rows_B,&cols_B,&mb,&nb,&IZERO,&IZERO,&nodes[i]->context,&idummy,&ierr);
      pgemm(Trans,'N',rows_z1,d,rows_y2,ONE,Trans=='N'?nodes[i]->B12:nodes[i]->B21,IONE,IONE,descGen,tmp2,rows_y1+1,IONE,desc,ZERO,tmp,IONE,IONE,descZ);

      /* Child 2: z{child2}=B21{i}*y{child1} */
      rows_B=Trans=='N'?child2->u_rank:child1->u_rank;
      cols_B=Trans=='N'?child1->v_rank:child2->v_rank;
      idummy=std::max(1,numroc_(&rows_B,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow));
      descinit_(descGen,&rows_B,&cols_B,&mb,&nb,&IZERO,&IZERO,&nodes[i]->context,&idummy,&ierr);
      pgemm(Trans,'N',rows_z2,d,rows_y1,ONE,Trans=='N'?nodes[i]->B21:nodes[i]->B12,IONE,IONE,descGen,tmp2,IONE,IONE,desc,ZERO,tmp,rows_z1+1,IONE,descZ);

      delete[] tmp2;

      /* Child 1 and 2: if !root, [z{child1};z{child2}]+=U{i}*z{i} */
      if(i!=last) {
        desc=new int[BLACSCTXTSIZE];
        rows_z=Trans=='N'?nodes[i]->u_rank:nodes[i]->v_rank;
        idummy=std::max(1,numroc_(&rows_z,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow));
        descinit_(desc,&rows_z,&d,&mb,&nb,&IZERO,&IZERO,&nodes[i]->context,&idummy,&ierr);
        nodes[i]->par_applyGenerator('N',Trans=='N'?'U':'V',d,ONE,z[i],desc,ONE,tmp,descZ,&tmpinfo);
        delete[] desc;
      }
    } else {
      tmp=NULL;
      descZ=new int[BLACSCTXTSIZE];
      descset_(descZ,&br,&d,&mb,&nb,&IZERO,&IZERO,&INONE,&IONE);
    }

    /* Redistribute */
    tmpinfo1.ctxt=child1->context;
    tmpinfo1.nprow=child1->nprow;
    tmpinfo1.npcol=child1->npcol;
    tmpinfo1.myrow=child1->myrow;
    tmpinfo1.mycol=child1->mycol;
    tmpinfo1.mb=ginfo->mb;
    tmpinfo1.nb=ginfo->nb;
    tmpinfo2.ctxt=child2->context;
    tmpinfo2.nprow=child2->nprow;
    tmpinfo2.npcol=child2->npcol;
    tmpinfo2.myrow=child2->myrow;
    tmpinfo2.mycol=child2->mycol;
    tmpinfo2.mb=ginfo->mb;
    tmpinfo2.nb=ginfo->nb;

    splitVector(tmp,child1->procs[0],child1->procs[0]+child1->nprow*child1->npcol-1,rows_z1,child2->procs[0],child2->procs[0]+child2->nprow*child2->npcol-1,rows_z2,d,descZ,&tmp2,&desc,nodes[i]->allctxt,&tmpinfo1,&tmpinfo2,comm);
    delete[] tmp;
     

    if(myid<child1->procs[0]+child1->nprow*child1->npcol) {
      z[ch1]=tmp2;
    } else if(myid>=child2->procs[0] && myid<child2->procs[0]+child2->nprow*child2->npcol) {
      z[ch2]=tmp2;
    }
    delete[] descZ;
    descZ=desc;
  }

  /* Serial subtree */
  if(subtree[2*myid]>=first) {
    istart=subtree[2*myid];
    iend=subtree[2*myid+1];
    ldl=np==1?a_dim:nodes[iend]->tree_ie-nodes[iend]->tree_ib+1;
    for(i=iend;i>=istart;i--) {
      ib=nodes[i]->tree_ib-nodes[istart]->tree_ib;
      b=nodes[i]->tree_ie-nodes[i]->tree_ib+1;
      rows_z=Trans=='N'?nodes[i]->u_rank:nodes[i]->v_rank;
      if(nodes[i]->nr_children==0) {
        /* Leaf: b(I,:)=U{i}*z{i}+D{i}*x(I,:) */
        gemm(Trans,'N',b,d,b,alpha,nodes[i]->D,b,&Xl[ib],ldl,beta,&Bl[ib],ldl);
        if(i!=last)
          /* Avoid weird case leaf=root (1 node in the tree) */
          nodes[i]->seq_applyGenerator('N',Trans=='N'?'U':'V',d,ONE,z[i],rows_z,ONE,&Bl[ib],ldl);
      } else {
        /* Non-leaf */
        ch1=children1[i];
        ch2=children2[i];
        child1=nodes[i]->children[0];
        child2=nodes[i]->children[1];
        rows_z1=Trans=='N'?child1->u_rank:child1->v_rank;
        rows_z2=Trans=='N'?child2->u_rank:child2->v_rank;
        rows_y1=Trans=='N'?child1->v_rank:child1->u_rank;
        rows_y2=Trans=='N'?child2->v_rank:child2->u_rank;
        /* Child 1: z{child1}=B12{i}*y{child2}+U{i}(1:r1)*z{i}; first part */
        z[ch1]=new T[rows_z1*d]();
        if(rows_z1 && rows_y2)
          gemm(Trans,'N',rows_z1,d,rows_y2,ONE,Trans=='N'?nodes[i]->B12:nodes[i]->B21,Trans=='N'?child1->u_rank:child2->u_rank,y[ch2],rows_y2,ZERO,z[ch1],rows_z1);

        /* Child 2: z{child2}=B21{i}*y{child1}+U{i}(r1+1:r1+r2)*z{i}; first part */
        z[ch2]=new T[rows_z2*d]();
        if(rows_z2 && rows_y1)
          gemm(Trans,'N',rows_z2,d,rows_y1,ONE,Trans=='N'?nodes[i]->B21:nodes[i]->B12,Trans=='N'?child2->u_rank:child1->u_rank,y[ch1],rows_y1,ZERO,z[ch2],rows_z2);

        /* Child 1 + Child 2; second part */
        if(i!=last) {
          tmp=new T[(rows_z1+rows_z2)*d]();
          nodes[i]->seq_applyGenerator('N',Trans=='N'?'U':'V',d,ONE,z[i],rows_z,ZERO,tmp,rows_z1+rows_z2);
          for(j=0;j<d;j++) {
            for(k=0;k<rows_z1;k++)
              z[ch1][j*rows_z1+k]+=tmp[j*(rows_z1+rows_z2)+k];
            for(k=0;k<rows_z2;k++)
              z[ch2][j*rows_z2+k]+=tmp[j*(rows_z1+rows_z2)+k+rows_z1];
          }
          delete[] tmp;
        }
      }
    }
  }

  /* Leaves */
  for(i=first;i<=last;i++) {
    if(   nodes[i]->procs[1]>nodes[i]->procs[0]
       && nodes[i]->nr_children==0
       && nodes[i]->procs[0]+nodes[i]->nprow*nodes[i]->npcol-1>=myid
       && nodes[i]->procs[0]<=myid
      ) {
      tmpinfo.ctxt=nodes[i]->context;
      tmpinfo.nprow=nodes[i]->nprow;
      tmpinfo.npcol=nodes[i]->npcol;
      tmpinfo.myrow=nodes[i]->myrow;
      tmpinfo.mycol=nodes[i]->mycol;
      tmpinfo.mb=ginfo->mb;
      tmpinfo.nb=ginfo->nb;
      rows_z=Trans=='N'?nodes[i]->u_rank:nodes[i]->v_rank;
      /* Leaf: b(I,:)=U{i}*z{i}+D{i}*x(I,:) */
      b=nodes[i]->tree_ie-nodes[i]->tree_ib+1;
      idummy=std::max(1,numroc_(&b,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow));
      descinit_(descGen,&b,&b,&mb,&nb,&IZERO,&IZERO,&nodes[i]->context,&idummy,&ierr);
      pgemm(Trans,'N',b,d,b,alpha,nodes[i]->D,IONE,IONE,descGen,Xl,IONE,IONE,descBX,beta,Bl,IONE,IONE,descBX);
      if(i!=last) {
        idummy=std::max(1,numroc_(&rows_z,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow));
        descZ=new int[BLACSCTXTSIZE];
        descinit_(descZ,&rows_z,&d,&mb,&nb,&IZERO,&IZERO,&nodes[i]->context,&idummy,&ierr);
        nodes[i]->par_applyGenerator('N',Trans=='N'?'U':'V',d,ONE,z[i],descZ,ONE,Bl,descBX,&tmpinfo);
        delete[] descZ;
      }
    }
  }

  /* Redistribute Bl in 2D block-cyclic form */
  undistVectorTree(Bl,descBX,d,B,descB,ginfo,nodes[dim_tree-1]->allctxt,partial);

  delete[] Xl;
  delete[] Bl;

  for(i=0;i<dim_tree-1;i++) {
    delete[] y[i];
    delete[] z[i];
  }
  delete[] y;
  delete[] z;

}

#endif // HSS_PAR_PRODUCT_HPP
