/*
 * STRUMPACK -- STRUctured Matrices PACKage, Copyright (c) 2014, The Regents of
 * the University of California, through Lawrence Berkeley National Laboratory
 * (subject to receipt of any required approvals from the U.S. Dept. of Energy).
 * All rights reserved.
 *
 * If you have questions about your rights to use or distribute this software,
 * please contact Berkeley Lab's Technology Transfer Department at TTD@lbl.gov.
 *
 * NOTICE. This software is owned by the U.S. Department of Energy. As such, the
 * U.S. Government has been granted for itself and others acting on its behalf a
 * paid-up, nonexclusive, irrevocable, worldwide license in the Software to
 * reproduce, prepare derivative works, and perform publicly and display publicly.
 * Beginning five (5) years after the date permission to assert copyright is
 * obtained from the U.S. Department of Energy, and subject to any subsequent five
 * (5) year renewals, the U.S. Government is granted for itself and others acting
 * on its behalf a paid-up, nonexclusive, irrevocable, worldwide license in the
 * Software to reproduce, prepare derivative works, distribute copies to the
 * public, perform publicly and display publicly, and to permit others to do so.
 *
 * Developers: Francois-Henry Rouet, Xiaoye S. Li, Pieter Ghysels
 *             (Lawrence Berkeley National Lab, Computational Research Division).
 *
 */

#ifndef HSS_PAR_ULV_HPP
#define HSS_PAR_ULV_HPP

template<typename T, typename S> void HSS_tree<T,S>::HSS_par_ULV(gridinfo *ginfo, int partial) {
  /* Parallel ULV-like factorization of an HSS matrix.
   *
   * The parallelization strategy is the same as for the compression.
   * A basic proportional mapping is applied to the HSS tree. Each process owns
   * a serial subtree (which might be empty) and a chain of nodes that are
   * shared with other processes. We call "parallel leaves":
   *  - The root nodes of the serial subtrees (nodes mapped on one process
   *  with a parent mapped on two processes or more).
   *  - The leaves that happen to be mapped on more than one process (nodes
   *  mapped on two processes or more without a node below them).
   * In other words, "parallel leaves" are the leaves of the parallel part of
   * the tree). The set of parallel leaves induces a clustering of [1 n].
   * In terms of implementation, we know that a node is a parallel leaf when
   *      procs[1]==procs[0] and nprow>0 (nodes inside serial subtrees have nprow=-1)
   *   or procs[1]> procs[0] and nr_children=0.
   * It is important to distinguish between two different things:
   *   - the set of processes an node is mapped on.
   *   - the set of processes that *work* at a node.
   * The difference comes from the fact that we use a BLACS grid at each node
   * and we try to maintain squarish grids. For example, if a node is mapped on
   * 37 processes, then only 36 actually belong to the (6x6) grid and own
   * pieces of of the data associated with the node. The 37th process is idle,
   * but it may be active at some descendants or ancestors of the node. The
   * code is as general as it can be.
   *
   * When "partial" is nonzero, only the lefmost subtree (rooted at the
   * "left"/"first" child of the root node) is factored. This works only
   * when the tree has 2 levels or more.
   */

  int mb, nb;
  int locr, locc;
  int idummy,ierr;
  int descVe[BLACSCTXTSIZE], descD[BLACSCTXTSIZE], desc[BLACSCTXTSIZE], descVt1[BLACSCTXTSIZE], descVt2[BLACSCTXTSIZE];
  int i, loc_i, j, k;
  int istart, iend;
  int br, bc, br1, br2;
  int myid, np;
  int proc_b, proc_g, proc_e, proc_bc, proc_gc;
  int last;
  T *Ve, *Vt1, *Vt2, *tmp;
  T *tau;
  HSS_tree<T,S> loc_htree;
  HSS_node<T,S> *child1, *child2;

  mb=ginfo->mb;
  nb=ginfo->nb;

  MPI_Comm_rank(comm,&myid);
  MPI_Comm_size(comm,&np);

  if(partial) {
    /* Searching for the index of the left child of the root node. */
    for(last=0;last<dim_tree;last++) {
      if(nodes[last]==nodes[dim_tree-1]->children[0])
        break;
    }
    knode=last;
  } else {
    last=dim_tree-1;
  }

  /* First, each process traverses its serial subtree.
   *
   * A local HSS tree is created by taking only nodes subtree[2*myid]...
   * subtree[2*myid+1] of htree.
   *
   * If there are more than one process, the sequential routine is called for a
   * subtree of the complete tree. At the root node of this subtree, LQ and
   * other operations must be performed, which is not the case when the
   * sequential subtree is equal to whole tree. By calling with doRoot=np-1, we
   * ensure that the root node of the serial subtree is correctly processed
   * (LU if np==1, LQ if np>1).
   */
  if(subtree[2*myid]>=0 && (!partial || subtree[2*myid+1]<=knode)) {
    /* Partial factorization: traverse only subtrees below
     * the left child of the root node.
     */
    extractLocalSubtree(&loc_htree);
    /* Copy compression information from global tree to local subtree */
    istart=subtree[2*myid];
    iend=subtree[2*myid+1];
    for(i=istart,loc_i=0;i<=iend;i++,loc_i++) {
      loc_htree.nodes[loc_i]->U_E=nodes[i]->U_E;
      loc_htree.nodes[loc_i]->U_perm=nodes[i]->U_perm;
      loc_htree.nodes[loc_i]->u_rank=nodes[i]->u_rank;
      loc_htree.nodes[loc_i]->u_rows=nodes[i]->u_rows;
      loc_htree.nodes[loc_i]->V_E=nodes[i]->V_E;
      loc_htree.nodes[loc_i]->V_perm=nodes[i]->V_perm;
      loc_htree.nodes[loc_i]->v_rank=nodes[i]->v_rank;
      loc_htree.nodes[loc_i]->v_rows=nodes[i]->v_rows;
      loc_htree.nodes[loc_i]->B12=nodes[i]->B12;
      loc_htree.nodes[loc_i]->B21=nodes[i]->B21;
      loc_htree.nodes[loc_i]->D=nodes[i]->D;
    }
    /* Run the serial factorization on the subtree */
    if(partial) {
      if(iend==dim_tree-1) {
        /* The serial subtree is the whole tree. We run HSS_seq_ULV
         * with doRoot=0 (special treatment of the root node),
         * partial=1 (we traverse only the "left" part); saveVhat
         * doesn't matter here.
         */
        loc_htree.HSS_seq_ULV(0,1,0);
      } else if(iend==knode) {
        /* The subtree is rooted at knode, i.e., it is
         * the one we want to factor. We run HSS_seq_ULV
         * asking for a complete factorization of the subtree
         * (doRoot=0, partial=0), but we save Vhat_k (saveVhat=1).
         */
        loc_htree.HSS_seq_ULV(0,0,1);
      } else {
        /* The subtree is rooted below knode. We run HSS_seq_ULV
         * asking for a complete factorization where the root node
         * of the subtree is treated as any other node (doRoot=1, partial=0)
         * and without saving Vhat at the root of the subtree.
         */
        loc_htree.HSS_seq_ULV(1,0,0);
      }
    } else {
      /* Regular factorization: call sequential ULV on the subtree,
       * with doRoot=np-1: if serial, then the subtree is the
       * whole tree, so the root node is special; otherwise,
       * the root node of the subtree is treated as any other
       * node (doRoot!=0). partial=0 (we traverse the whole subtree)
       * and saveVhat=0 (no need to save Vhat at k).
       */
      loc_htree.HSS_seq_ULV(np-1,0,0);
    }

    /* Copy/point info from serial subtree to global tree */
    for(i=istart,loc_i=0;i<=iend;i++,loc_i++) {
      nodes[i]->D=loc_htree.nodes[loc_i]->D;
      nodes[i]->Dt=loc_htree.nodes[loc_i]->Dt;
      nodes[i]->L=loc_htree.nodes[loc_i]->L;
      nodes[i]->piv=loc_htree.nodes[loc_i]->piv;
      nodes[i]->Vt=loc_htree.nodes[loc_i]->Vt;
      nodes[i]->W=loc_htree.nodes[loc_i]->W;
      nodes[i]->Q=loc_htree.nodes[loc_i]->Q;
      if(i==dim_tree-1)
        /* Parallel code called in serial; need to copy pivoting info */
        nodes[i]->U_perm=loc_htree.nodes[loc_i]->U_perm;
    }
  }

  /* Second, parallel nodes are visited. This is the same algorithm as
   * in the lower part, except that arrays are constantly distributed
   * in a 2D block-cyclic format.
   */
  for(j=0;j<par_lvl+1;j++) {
    /* Same loop for leaves and interior nodes */
    if(j>=1) {
      /* Follow the chain of parallel interior nodes */
      i=par_nodes[j-1];
      if(i>last)
        break;
    } else {
      /* Find local parallel leaf (if any) */
      for(i=0;i<=last;i++) {
        proc_b=nodes[i]->procs[0];
        proc_g=nodes[i]->procs[0]+nodes[i]->nprow*nodes[i]->npcol-1;
        proc_e=nodes[i]->procs[1];
        if(proc_e>proc_b && myid>=proc_b && myid <= proc_g && nodes[i]->nr_children==0)
          /* Leaf mapped on several processes on which myid works */
          break;
      }
      if(i>last)
       /* No parallel leaf, go to interior nodes */
       continue;
    }

    /* Processing node i */
    if(nodes[i]->nr_children==0) {
      br=nodes[i]->tree_ie-nodes[i]->tree_ib+1;
      bc=nodes[i]->tree_ie-nodes[i]->tree_ib+1;
    } else {
      child1=nodes[i]->children[0];
      child2=nodes[i]->children[1];
      br=child1->u_rank+child2->u_rank;
      bc=child1->v_rank+child2->v_rank;
      br1=child1->u_rows+child1->u_rank;
      br2=child2->u_rows+child2->u_rank;
    }
    /* Node i is mapped on proc_b...proc_e but only
     * proc_b...proc_g are active (in the grid).
     */
    proc_b=nodes[i]->procs[0];
    proc_g=nodes[i]->procs[0]+nodes[i]->nprow*nodes[i]->npcol-1;
    proc_e=nodes[i]->procs[1];

    if(i!=dim_tree-1 && myid>=proc_b && myid<=proc_g) {
      /* We expand V into its explicit form V=P*[I;E] */
      locr=numroc_(&bc,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow);
      locc=numroc_(&nodes[i]->v_rank,&nb,&nodes[i]->mycol,&IZERO,&nodes[i]->npcol);
      Ve=new T[locr*locc]();
      idummy=std::max(1,locr);
      descinit_(descVe,&bc,&nodes[i]->v_rank,&mb,&nb,&IZERO,&IZERO,&nodes[i]->context,&idummy,&ierr);
      for(k=1;k<=nodes[i]->v_rank;k++)
        if(nodes[i]->myrow==indxg2p_(&k,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow) && nodes[i]->mycol==indxg2p_(&k,&nb,&nodes[i]->mycol,&IZERO,&nodes[i]->npcol))
          Ve[(indxg2l_(&k,&nb,&nodes[i]->mycol,&IZERO,&nodes[i]->npcol)-1)*locr+indxg2l_(&k,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow)-1]=ONE;
      pgeadd('N',nodes[i]->v_rows,nodes[i]->v_rank,ONE,nodes[i]->V_E,IONE,IONE,nodes[i]->descV,ZERO,Ve,nodes[i]->v_rank+1,IONE,descVe);
      for(k=bc;k>=1;k--)
        if(k!=nodes[i]->V_perm[k-1])
          pswap(nodes[i]->v_rank,Ve,k,IONE,descVe,bc,Ve,nodes[i]->V_perm[k-1],IONE,descVe,bc);
    }

    if(nodes[i]->nr_children!=0) {
      /* D = [ Dt_1 B12*Vt_2(rows2+1:end)' ; B21*Vt_1(rows1+1:end)' Dt_2] */
      if(myid>=proc_b && myid<=proc_g) {
        locr=numroc_(&br,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow);
        locc=numroc_(&br,&nb,&nodes[i]->mycol,&IZERO,&nodes[i]->npcol);
        nodes[i]->D=new T[locr*locc];
        idummy=std::max(1,locr);
        descinit_(descD,&br,&br,&mb,&nb,&IZERO,&IZERO,&nodes[i]->context,&idummy,&ierr);
      } else
        descset_(descD,&br,&br,&mb,&nb,&IZERO,&IZERO,&INONE,&INONE);

      /* (1,1) block */
      proc_bc=child1->procs[0];
      proc_gc=child1->procs[0]+child1->nprow*child1->npcol-1;

      if(myid>=proc_bc && myid<=proc_gc) {
        locr=numroc_(&child1->u_rank,&mb,&child1->myrow,&IZERO,&child1->nprow);
        locc=numroc_(&child1->u_rank,&nb,&child1->mycol,&IZERO,&child1->npcol);
        idummy=std::max(1,locr);
        descinit_(desc,&child1->u_rank,&child1->u_rank,&mb,&nb,&IZERO,&IZERO,&child1->context,&idummy,&ierr);
      } else
        descset_(desc,&child1->u_rank,&child1->u_rank,&mb,&nb,&IZERO,&IZERO,&INONE,&INONE);

      pgemr2d(child1->u_rank,child1->u_rank,child1->Dt,IONE,IONE,desc,nodes[i]->D,IONE,IONE,descD,nodes[i]->allctxt);

      /* (1,2) block: redistribute/copy child2->Vt_2 into Vt2 then compute the product */
      proc_bc=child2->procs[0];
      proc_gc=child2->procs[0]+child2->nprow*child2->npcol-1;

      if(myid>=proc_bc && myid<=proc_gc) {
        locr=numroc_(&br2,&mb,&child2->myrow,&IZERO,&child2->nprow);
        locc=numroc_(&child2->v_rank,&nb,&child2->mycol,&IZERO,&child2->npcol);
        idummy=std::max(1,locr);
        descinit_(desc,&br2,&child2->v_rank,&mb,&nb,&IZERO,&IZERO,&child2->context,&idummy,&ierr);
      } else
        descset_(desc,&br2,&child2->v_rank,&mb,&nb,&IZERO,&IZERO,&INONE,&INONE);

      if(myid>=proc_b && myid<=proc_g) {
        locr=numroc_(&child2->u_rank,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow);
        locc=numroc_(&child2->v_rank,&nb,&nodes[i]->mycol,&IZERO,&nodes[i]->npcol);
        Vt2=new T[locr*locc];
        idummy=std::max(1,locr);
        descinit_(descVt2,&child2->u_rank,&child2->v_rank,&mb,&nb,&IZERO,&IZERO,&nodes[i]->context,&idummy,&ierr);
      } else
        descset_(descVt2,&child2->u_rank,&child2->v_rank,&mb,&nb,&IZERO,&IZERO,&INONE,&INONE);

      pgemr2d(child2->u_rank,child2->v_rank,child2->Vt,child2->u_rows+1,IONE,desc,Vt2,IONE,IONE,descVt2,nodes[i]->allctxt);

      if(myid>=proc_b && myid<=proc_g) {
        idummy=std::max(1,numroc_(&child1->u_rank,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow));
        descinit_(desc,&child1->u_rank,&child2->v_rank,&mb,&nb,&IZERO,&IZERO,&nodes[i]->context,&idummy,&ierr);
        pgemm('N','C',child1->u_rank,child2->u_rank,child2->v_rank,ONE,nodes[i]->B12,IONE,IONE,desc,Vt2,IONE,IONE,descVt2,ZERO,nodes[i]->D,IONE,child1->u_rank+1,descD);
      }

      /* (2,1) block: redistribute/copy child1->Vt_2 into Vt1 then compute the product */
      proc_bc=child1->procs[0];
      proc_gc=child1->procs[0]+child1->nprow*child1->npcol-1;

      if(myid>=proc_bc && myid<=proc_gc) {
        locr=numroc_(&br1,&mb,&child1->myrow,&IZERO,&child1->nprow);
        locc=numroc_(&child1->v_rank,&nb,&child1->mycol,&IZERO,&child1->npcol);
        idummy=std::max(1,locr);
        descinit_(desc,&br1,&child1->v_rank,&mb,&nb,&IZERO,&IZERO,&child1->context,&idummy,&ierr);
      } else
        descset_(desc,&br1,&child1->v_rank,&mb,&nb,&IZERO,&IZERO,&INONE,&INONE);

      if(myid>=proc_b && myid<=proc_g) {
        locr=numroc_(&child1->u_rank,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow);
        locc=numroc_(&child1->v_rank,&nb,&nodes[i]->mycol,&IZERO,&nodes[i]->npcol);
        Vt1=new T[locr*locc];
        idummy=std::max(1,locr);
        descinit_(descVt1,&child1->u_rank,&child1->v_rank,&mb,&nb,&IZERO,&IZERO,&nodes[i]->context,&idummy,&ierr);
      } else
        descset_(descVt1,&child1->u_rank,&child1->v_rank,&mb,&nb,&IZERO,&IZERO,&INONE,&INONE);

      pgemr2d(child1->u_rank,child1->v_rank,child1->Vt,child1->u_rows+1,IONE,desc,Vt1,IONE,IONE,descVt1,nodes[i]->allctxt);

      if(myid>=proc_b && myid<=proc_g) {
        idummy=std::max(1,numroc_(&child2->u_rank,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow));
        descinit_(desc,&child2->u_rank,&child1->v_rank,&mb,&nb,&IZERO,&IZERO,&nodes[i]->context,&idummy,&ierr);
        pgemm('N','C',child2->u_rank,child1->u_rank,child1->v_rank,ONE,nodes[i]->B21,IONE,IONE,desc,Vt1,IONE,IONE,descVt1,ZERO,nodes[i]->D,child1->u_rank+1,IONE,descD);
      }

      /* (2,2) block */
      proc_bc=child2->procs[0];
      proc_gc=child2->procs[0]+child2->nprow*child2->npcol-1;

      if(myid>=proc_bc && myid<=proc_gc) {
        locr=numroc_(&child2->u_rank,&mb,&child2->myrow,&IZERO,&child2->nprow);
        locc=numroc_(&child2->u_rank,&nb,&child2->mycol,&IZERO,&child2->npcol);
        idummy=std::max(1,locr);
        descinit_(desc,&child2->u_rank,&child2->u_rank,&mb,&nb,&IZERO,&IZERO,&child2->context,&idummy,&ierr);
      } else
        descset_(desc,&child2->u_rank,&child2->u_rank,&mb,&nb,&IZERO,&IZERO,&INONE,&INONE);

      pgemr2d(child2->u_rank,child2->u_rank,child2->Dt,IONE,IONE,desc,nodes[i]->D,child1->u_rank+1,child1->u_rank+1,descD,nodes[i]->allctxt);

      if(i!=dim_tree-1 && myid>=proc_b && myid<=proc_g) {
        /* V = [Vt_1(rows1+1:end) 0 ; 0 Vt_2(rows2+1:end) ] * V (in array Ve) */
        locr=numroc_(&br,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow);
        locc=numroc_(&nodes[i]->v_rank,&nb,&nodes[i]->mycol,&IZERO,&nodes[i]->npcol);
        tmp=new T[locr*locc];
        idummy=std::max(1,locr);
        descinit_(desc,&br,&nodes[i]->v_rank,&mb,&nb,&IZERO,&IZERO,&nodes[i]->context,&idummy,&ierr);
        /* Recall that Vt* contains only rows child*->u_rows+1:end of child*->Vt */
        pgemm('N','N',child1->u_rank,nodes[i]->v_rank,child1->v_rank,ONE,Vt1,IONE,IONE,descVt1,Ve,IONE,IONE,descVe,ZERO,tmp,IONE,IONE,desc);
        pgemm('N','N',child2->u_rank,nodes[i]->v_rank,child2->v_rank,ONE,Vt2,IONE,IONE,descVt2,Ve,child1->v_rank+1,IONE,descVe,ZERO,tmp,child1->u_rank+1,IONE,desc);

        delete[] Ve;
        Ve=tmp;
        descinit_(descVe,&br,&nodes[i]->v_rank,&mb,&nb,&IZERO,&IZERO,&nodes[i]->context,&idummy,&ierr);
      }
    } else {
      /* For leaves, simply define the descriptor for D */
      if(myid>=proc_b && myid<=proc_g) {
        locr=numroc_(&br,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow);
        locc=numroc_(&br,&nb,&nodes[i]->mycol,&IZERO,&nodes[i]->npcol);
        idummy=std::max(1,locr);
        descinit_(descD,&br,&br,&mb,&nb,&IZERO,&IZERO,&nodes[i]->context,&idummy,&ierr);
      }
    }

    if(partial && i==last) {
      /* Vhat (if k is an interior node) of V (if k is a leaf)
       * stored in Vt and used for Schur complement computation.
       * Note that the following step is the LU factorization
       * and Ve is not destroyed, so we can safely point to it.
       */
      nodes[i]->Vt=Ve;
    }


    if(myid>=proc_b && myid<=proc_g) {
      if(i==last) {
        /* Dense LU at root node */
        locr=numroc_(&br,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow);
        locc=numroc_(&br,&nb,&nodes[i]->mycol,&IZERO,&nodes[i]->npcol);
        nodes[i]->L=new T[locr*locc];
        idummy=std::max(1,locr);
        nodes[i]->descL=new int[BLACSCTXTSIZE];
        descinit_(nodes[i]->descL,&br,&br,&mb,&nb,&IZERO,&IZERO,&nodes[i]->context,&idummy,&ierr);
        lacpy('N',locr,locc,nodes[i]->D,locr,nodes[i]->L,locr);
        nodes[i]->piv=new int[locr+mb];
        ierr=pgetrf(br,br,nodes[i]->L,IONE,IONE,nodes[i]->descL,nodes[i]->piv);
        if(ierr!=0)
          std::cout << "Something went wrong with LU at root node" << std::endl;
      } else {
        /* W = [-U_E I ; I 0 ] * U_perm' * D */
        /* U_perm' * D is stored in Dt temporarily */
        locr=numroc_(&br,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow);
        locc=numroc_(&br,&nb,&nodes[i]->mycol,&IZERO,&nodes[i]->npcol);
        nodes[i]->Dt=new T[locr*locc];
        lacpy('N',locr,locc,nodes[i]->D,locr,nodes[i]->Dt,locr);
        for(k=1;k<=br;k++)
          if(k!=nodes[i]->U_perm[k-1])
            pswap(br,nodes[i]->Dt,k,IONE,descD,br,nodes[i]->Dt,nodes[i]->U_perm[k-1],IONE,descD,br);

        /* W = [ W1 ; W2 ] with W1 = Dt(u_rank+1:end,:)- U_E * Dt(1:u_rank,:) and W2 = Dt(1:u_rank,:) */
        nodes[i]->W=new T[locr*locc];
        idummy=std::max(1,locr);
        nodes[i]->descW=new int[BLACSCTXTSIZE];
        descinit_(nodes[i]->descW,&br,&br,&mb,&nb,&IZERO,&IZERO,&nodes[i]->context,&idummy,&ierr);

        if(nodes[i]->u_rows) {
          pgeadd('N',nodes[i]->u_rows,br,ONE,nodes[i]->Dt,nodes[i]->u_rank+1,IONE,descD,ZERO,nodes[i]->W,IONE,IONE,nodes[i]->descW);
          pgemm('N','N',nodes[i]->u_rows,br,nodes[i]->u_rank,NONE,nodes[i]->U_E,IONE,IONE,nodes[i]->descU,nodes[i]->Dt,IONE,IONE,descD,ONE,nodes[i]->W,IONE,IONE,nodes[i]->descW);

        }
        pgeadd('N',nodes[i]->u_rank,br,ONE,nodes[i]->Dt,IONE,IONE,descD,ZERO,nodes[i]->W,nodes[i]->u_rows+1,IONE,nodes[i]->descW);
        delete[] nodes[i]->Dt; nodes[i]->Dt=NULL;

        if(nodes[i]->u_rows) {
          /* LQ(W1): W1 = [L 0] * Q */
          locr=numroc_(&br,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow);
          locc=numroc_(&br,&nb,&nodes[i]->mycol,&IZERO,&nodes[i]->npcol);
          nodes[i]->descQ=new int[BLACSCTXTSIZE];
          idummy=std::max(1,locr);
          descinit_(nodes[i]->descQ,&br,&br,&mb,&nb,&IZERO,&IZERO,&nodes[i]->context,&idummy,&ierr);
          nodes[i]->Q=new T[locr*locc]();
          placpy('N',nodes[i]->u_rows,br,nodes[i]->W,IONE,IONE,nodes[i]->descW,nodes[i]->Q,IONE,IONE,nodes[i]->descQ);
          tau=new T[locr+locc];
          pgelqf(nodes[i]->u_rows,br,nodes[i]->Q,IONE,IONE,nodes[i]->descW,tau,mb,nb,nodes[i]->myrow,nodes[i]->mycol,nodes[i]->nprow,nodes[i]->npcol);

          /* Copy the L factor into nodes[i]->L; we don't discard the upper part */
          locr=numroc_(&nodes[i]->u_rows,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow);
          locc=numroc_(&nodes[i]->u_rows,&nb,&nodes[i]->mycol,&IZERO,&nodes[i]->npcol);
          nodes[i]->L=new T[locr*locc];
          idummy=std::max(1,locr);
          nodes[i]->descL=new int[BLACSCTXTSIZE];
          descinit_(nodes[i]->descL,&nodes[i]->u_rows,&nodes[i]->u_rows,&mb,&nb,&IZERO,&IZERO,&nodes[i]->context,&idummy,&ierr);
          pgeadd('N',nodes[i]->u_rows,nodes[i]->u_rows,ONE,nodes[i]->Q,IONE,IONE,nodes[i]->descQ,ZERO,nodes[i]->L,IONE,IONE,nodes[i]->descL);

          /* Form the explicit Q factor from the Householder reflectors */
          pxxglq(br,br,nodes[i]->u_rows,nodes[i]->Q,IONE,IONE,nodes[i]->descQ,tau,mb,nb,nodes[i]->myrow,nodes[i]->mycol,nodes[i]->nprow,nodes[i]->npcol);
          delete[] tau;

          /* Vt = Q V ; V (explicit Ve) is not used afterwards */
          locr=numroc_(&br,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow);
          locc=numroc_(&nodes[i]->v_rank,&nb,&nodes[i]->mycol,&IZERO,&nodes[i]->npcol);
          nodes[i]->Vt=new T[locr*locc];
          idummy=std::max(1,locr);
          nodes[i]->descVt=new int[BLACSCTXTSIZE];
          descinit_(nodes[i]->descVt,&br,&nodes[i]->v_rank,&mb,&nb,&IZERO,&IZERO,&nodes[i]->context,&idummy,&ierr);
          pgemm('N','N',br,nodes[i]->v_rank,br,ONE,nodes[i]->Q,IONE,IONE,nodes[i]->descQ,Ve,IONE,IONE,descVe,ZERO,nodes[i]->Vt,IONE,IONE,nodes[i]->descVt);

          /* Dt = W(u_rows+1:end,:) * Q(u_rows+1:end,:)' */
          locr=numroc_(&nodes[i]->u_rank,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow);
          locc=numroc_(&nodes[i]->u_rank,&nb,&nodes[i]->mycol,&IZERO,&nodes[i]->npcol);
          nodes[i]->Dt=new T[locr*locc];
          idummy=std::max(1,locr);
          descinit_(descD,&nodes[i]->u_rank,&nodes[i]->u_rank,&mb,&nb,&IZERO,&IZERO,&nodes[i]->context,&idummy,&ierr);
          pgemm('N','C',nodes[i]->u_rank,nodes[i]->u_rank,br,ONE,nodes[i]->W,nodes[i]->u_rows+1,IONE,nodes[i]->descW,nodes[i]->Q,nodes[i]->u_rows+1,IONE,nodes[i]->descQ,ZERO,nodes[i]->Dt,IONE,IONE,descD);
        } else {
          /* If u_rows=0, Vt=V, D=W, Q=I (not stored) */
          locr=numroc_(&nodes[i]->u_rank,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow);
          locc=numroc_(&nodes[i]->v_rank,&nb,&nodes[i]->mycol,&IZERO,&nodes[i]->npcol);
          nodes[i]->Vt=new T[locr*locc];
          idummy=std::max(1,locr);
          nodes[i]->descVt=new int[BLACSCTXTSIZE];
          descinit_(nodes[i]->descVt,&nodes[i]->u_rank,&nodes[i]->v_rank,&mb,&nb,&IZERO,&IZERO,&nodes[i]->context,&idummy,&ierr);
          pgeadd('N',nodes[i]->u_rank,nodes[i]->v_rank,ONE,Ve,IONE,IONE,descVe,ZERO,nodes[i]->Vt,IONE,IONE,nodes[i]->descVt);

          locr=numroc_(&nodes[i]->u_rank,&mb,&nodes[i]->myrow,&IZERO,&nodes[i]->nprow);
          locc=numroc_(&nodes[i]->u_rank,&nb,&nodes[i]->mycol,&IZERO,&nodes[i]->npcol);
          nodes[i]->Dt=new T[locr*locc];
          idummy=std::max(1,locr);
          descinit_(descD,&nodes[i]->u_rank,&nodes[i]->u_rank,&mb,&nb,&IZERO,&IZERO,&nodes[i]->context,&idummy,&ierr);
          pgeadd('N',nodes[i]->u_rank,nodes[i]->u_rank,ONE,nodes[i]->W,IONE,IONE,nodes[i]->descW,ZERO,nodes[i]->Dt,IONE,IONE,descD);

          /* If u_rows=0, Vt=V, D=W, Q=I (not stored) */
        }
      }
    }
  }

  /* Clean-up: D for non-leaves and Dt are not needed anymore */
  for(i=0;i<dim_tree;i++) {
    if(nodes[i]->nr_children!=0) {
      delete[] nodes[i]->D; nodes[i]->D=NULL;
    }
    delete[] nodes[i]->Dt; nodes[i]->Dt=NULL;
  }
}

#endif // HSS_PAR_ULV_HPP
