/*
 * STRUMPACK -- STRUctured Matrices PACKage, Copyright (c) 2014, The Regents of
 * the University of California, through Lawrence Berkeley National Laboratory
 * (subject to receipt of any required approvals from the U.S. Dept. of Energy).
 * All rights reserved.
 *
 * If you have questions about your rights to use or distribute this software,
 * please contact Berkeley Lab's Technology Transfer Department at TTD@lbl.gov.
 *
 * NOTICE. This software is owned by the U.S. Department of Energy. As such, the
 * U.S. Government has been granted for itself and others acting on its behalf a
 * paid-up, nonexclusive, irrevocable, worldwide license in the Software to
 * reproduce, prepare derivative works, and perform publicly and display publicly.
 * Beginning five (5) years after the date permission to assert copyright is
 * obtained from the U.S. Department of Energy, and subject to any subsequent five
 * (5) year renewals, the U.S. Government is granted for itself and others acting
 * on its behalf a paid-up, nonexclusive, irrevocable, worldwide license in the
 * Software to reproduce, prepare derivative works, distribute copies to the
 * public, perform publicly and display publicly, and to permit others to do so.
 *
 * Developers: Francois-Henry Rouet, Xiaoye S. Li, Pieter Ghysels
 *             (Lawrence Berkeley National Lab, Computational Research Division).
 *
 */

#ifndef HSS_SEQ_SOLVE_HPP
#define HSS_SEQ_SOLVE_HPP

template<typename T, typename S> void HSS_tree<T,S>::HSS_seq_Solve(T *X, int d, T *B, int partial) {
  /* Solution of A*X=B using the ULV factors of A.
   * B and X have d columns.
   * "partial":
   *   - If 0, do the entire solve.
   *   - If 1, do only the forward pass.
   *   - If 2, do only the backward pass.
   */


  int i, j, k, br, bc, br1, br2, ib;
  int info;
  int last;
  int *children1, *children2;
  T *tmp;
  T **f, **x, **z;
  int ch1, ch2;
  HSS_node<T,S> *child1, *child2;

  /* Here f, x, y, are not fields of a node but arrays indexed following the
   * postorder; the pointers "children" of a node are not enough to find the
   * right in y and z, what we need is numbers in the postorder. "children1"
   * and "children2" are the same information as children but they provide
   * numbers (indices in the tree).
   *
   * children1: children[i]==j: node number j is left child of node number i.
   * child1: pointer to left child (type HSS_node) of node nodes[i].
   *
   * When partial==1: the "y" structure (field of the HSS tree) is saved at
   * the end of the forward pass (bottom-up) in order to be reused in the
   * backward pass (partial==2). With partial=1, [z1;z2] at the root node
   * is saved into zk12 (field of the tree).
   */

  children1=new int[dim_tree];
  for(i=0;i<dim_tree;i++) {
    children1[i]=-1;
    if(nodes[i]->nr_children!=0)
      for(j=0;j<dim_tree;j++)
        if(nodes[j]==nodes[i]->children[0]) {
          children1[i]=j;
          break;
        }
  }
  children2=new int[dim_tree];
  for(i=0;i<dim_tree;i++) {
    children2[i]=-1;
    if(nodes[i]->nr_children!=0)
      for(j=0;j<dim_tree;j++)
        if(nodes[j]==nodes[i]->children[1]) {
          children2[i]=j;
          break;
        }
  }

  if(partial) {
    /* Traverse only the subtree rooted at the left child
     * of the root node.
     */
    last=knode;
  } else {
    /* Traverse the whole tree. */
    last=dim_tree-1;
  }

  if(partial!=2) {
    f=new T*[dim_tree];
    for(i=0;i<dim_tree;i++)
      f[i]=NULL;
    y=new T*[dim_tree];
    for(i=0;i<dim_tree;i++)
      y[i]=NULL;
    z=new T*[dim_tree];
    for(i=0;i<dim_tree;i++)
      z[i]=NULL;

    /* Bottom-up traversal */
    for(i=0;i<=last;i++) {
      ib=nodes[i]->tree_ib-1;
      if(nodes[i]->nr_children==0) {
        br=nodes[i]->tree_ie-nodes[i]->tree_ib+1;
        bc=nodes[i]->tree_ie-nodes[i]->tree_ib+1;
      } else {
        child1=nodes[i]->children[0];
        child2=nodes[i]->children[1];
        br=child1->u_rank+child2->u_rank;
        bc=child1->v_rank+child2->v_rank;
        br1=child1->u_rows+child1->u_rank;
        br2=child2->u_rows+child2->u_rank;
        ch1=children1[i];
        ch2=children2[i];
      }

      if(nodes[i]->nr_children!=0) {
        /* f{i} = [ ft{child 1}_2 - W{child 1}_2*Q{child 1}_1'*y{child 1} - B12*z{child 2} ]
         *        [ ft{child 2}_2 - B12*z{child 1} - W{child 2}_2*Q{child 2}_1'*y{child 2} ]
         *      = [ (1,1) - (1,2) - (1,3) ]
         *        [ (2,1) - (2,2) - (2,3) ]
         */
        f[i]=new T[br*d];

        /* (1,1) */
        lacpy('N',child1->u_rank,d,&f[ch1][child1->u_rows],br1,f[i],br);

        /* (1,2) if child1->u_rows!=0 (otherwise y{child 1} empty) */
        if(child1->u_rows) {
          /* Q*y */
          tmp=new T[br1*d];
          if(br1 && child1->u_rows)
            gemm('C','N',br1,d,child1->u_rows,ONE,child1->Q,br1,y[ch1],child1->u_rows,ZERO,tmp,br1);

          /* W*Q*y */
          if(br && br1)
            gemm('N','N',child1->u_rank,d,br1,NONE,&child1->W[child1->u_rows],br1,tmp,br1,ONE,f[i],br);
          delete[] tmp;
        }

        /* (1,3) if child2->u_rows!=0 or non-leaf (otherwise z{child 2} 0) */
        if(child2->u_rows||child2->nr_children)
          if(child1->u_rank && child2->v_rank && br)
            gemm('N','N',child1->u_rank,d,child2->v_rank,NONE,nodes[i]->B12,child1->u_rank,z[ch2],child2->v_rank,ONE,f[i],br);

        /* (2,1) */
        lacpy('N',child2->u_rank,d,&f[ch2][child2->u_rows],br2,&f[i][child1->u_rank],br);

        /* (2,2) if child1->u_rows!=0 or non-leaf (otherwise z{child 1} 0) */
        if(child1->u_rows||child1->nr_children)
          if(child2->u_rank && child1->v_rank && br)
            gemm('N','N',child2->u_rank,d,child1->v_rank,NONE,nodes[i]->B21,child2->u_rank,z[ch1],child1->v_rank,ONE,&f[i][child1->u_rank],br);

        /* (2,3): if child2->u_rows!=0 (otherwise y{child 2} empty) */
        if(child2->u_rows) {
          /* Q*y */
          tmp=new T[br2*d];
          if(br2)
            gemm('C','N',br2,d,child2->u_rows,ONE,child2->Q,br2,y[ch2],child2->u_rows,ZERO,tmp,br2);

          /* W*Q*y */
          if(br && br2)
            gemm('N','N',child2->u_rank,d,br2,NONE,&child2->W[child2->u_rows],br2,tmp,br2,ONE,&f[i][child1->u_rank],br);
          delete[] tmp;
        }

      } else {
        f[i]=new T[br*d];
        lacpy('N',br,d,&B[ib],a_dim,f[i],br);
      }

      if(i==last) {
        /* Triangular solve with traditional LU factors at root */
        if(br>0) {
          info=getrs('N',br,d,nodes[i]->L,br,nodes[i]->piv,f[i],br);
          if(info!=0)
            std::cout << "Something went wrong with triangular solve at root node" << std::endl;
          y[i]=new T[br*d];
          lacpy('N',br,d,f[i],br,y[i],br);
        }
        if(partial==1 && nodes[i]->nr_children) {
          /* Save [z1;z2] into zk12 (field of the tree); will be used for RHS reduction */
          zk12=new T[bc*d];
          lacpy('N',child1->v_rank,d,z[ch1],child1->v_rank,zk12,bc);
          lacpy('N',child2->v_rank,d,z[ch2],child2->v_rank,&zk12[child1->v_rank],bc);
        }
      } else {

        /* ft{i} = [-U_E I ; 0 I ] * U_perm' * f{i} */
        /* f{i} is overwritten */
        for(k=0;k<br;k++)
          if(k!=nodes[i]->U_perm[k]-1)
            swap(d,&f[i][k],br,&f[i][nodes[i]->U_perm[k]-1],br);

        if(nodes[i]->u_rows) {
          /* Otherwise ft{i} = U_perm' * f{i}, y{i} empty */

          /* ft{i} = [ft{i}_1 ; ft{i}_2] with ft{i}_1=f{i}(u_rank+1:end,:)-U_E*f{i}(1:u_rank,:) and ft{i}_2=f{i}(1:u_rank,:) */
          tmp=new T[br*d]();
          lacpy('N',nodes[i]->u_rows,d,&f[i][nodes[i]->u_rank],br,tmp,br);
          gemm('N','N',nodes[i]->u_rows,d,nodes[i]->u_rank,NONE,nodes[i]->U_E,nodes[i]->u_rows,f[i],br,ONE,tmp,br);
          lacpy('N',nodes[i]->u_rank,d,f[i],br,&tmp[nodes[i]->u_rows],br);
          delete[] f[i];
          f[i]=tmp;

          /* y{i} = L{i}\ft{i}_1 */
          y[i]=new T[nodes[i]->u_rows*d];
          lacpy('N',nodes[i]->u_rows,d,f[i],br,y[i],nodes[i]->u_rows);
          trsm('L','L','N','N',nodes[i]->u_rows,d,ONE,nodes[i]->L,nodes[i]->u_rows,y[i],nodes[i]->u_rows);
        }
        /* z{i} = Vt{i}_1' * y{i} + V{i}' * [z{child 1} ; z{child 2}] (non leaves) */
        z[i]=new T[nodes[i]->v_rank*d]();
        if(nodes[i]->u_rows)
          if(br && nodes[i]->u_rows && nodes[i]->v_rank)
            gemm('C','N',nodes[i]->v_rank,d,nodes[i]->u_rows,ONE,nodes[i]->Vt,br,y[i],nodes[i]->u_rows,ZERO,z[i],nodes[i]->v_rank);
        if(nodes[i]->nr_children!=0) {
          tmp=new T[bc*d];
          lacpy('N',child1->v_rank,d,z[ch1],child1->v_rank,tmp,bc);
          lacpy('N',child2->v_rank,d,z[ch2],child2->v_rank,&tmp[child1->v_rank],bc);
          nodes[i]->seq_applyGenerator('C','V',d,ONE,tmp,bc,ONE,z[i],nodes[i]->v_rank);
          delete[] tmp;
        }
      }
    }

    for(i=0;i<dim_tree-1;i++) {
      delete[] f[i];
      delete[] z[i];
    }
    delete[] f;
    delete[] z;
  }

  if(partial!=1) {
    x=new T*[dim_tree];
    for(i=0;i<dim_tree;i++)
      x[i]=NULL;

    /* Top-down traversal */
    for(i=last;i>=0;i--) {
      ib=nodes[i]->tree_ib-1;
      if(nodes[i]->nr_children==0) {
        br=nodes[i]->tree_ie-nodes[i]->tree_ib+1;
        bc=nodes[i]->tree_ie-nodes[i]->tree_ib+1;
      } else {
        child1=nodes[i]->children[0];
        child2=nodes[i]->children[1];
        br=child1->u_rank+child2->u_rank;
        bc=child1->v_rank+child2->v_rank;
        br1=child1->u_rows+child1->u_rank;
        br2=child2->u_rows+child2->u_rank;
        ch1=children1[i];
        ch2=children2[i];
      }

      /* Partial solution at root is stored in y */
      if(i==last) {
        x[i]=new T[br*d];
        lacpy('N',br,d,y[i],br,x[i],br);
      }

      if(nodes[i]->nr_children!=0) {
        /* Child 1: x{child1} = Q{child1}'*[y{child1};x{i}_1] */
        if(child1->u_rows) {
          tmp=new T[br1*d]();
          lacpy('N',child1->u_rows,d,y[ch1],child1->u_rows,tmp,br1);
          lacpy('N',child1->u_rank,d,x[i],br,&tmp[child1->u_rows],br1);
          x[ch1]=new T[br1*d];
          gemm('C','N',br1,d,br1,ONE,child1->Q,br1,tmp,br1,ZERO,x[ch1],br1);
          delete[] tmp;
        } else {
          /* Q{child 1} = I (not stored), y{child 1} empty => x{child 1}=x{i}_1 */
          x[ch1]=new T[br1*d];
          lacpy('N',br1,d,x[i],br,x[ch1],br1);
        }

        /* Child 2: x{child2} = Q{child2}'*[y{child2};x{i}_2] */
        if(child2->u_rows) {
          tmp=new T[br2*d]();
          lacpy('N',child2->u_rows,d,y[ch2],child2->u_rows,tmp,br2);
          lacpy('N',child2->u_rank,d,&x[i][child1->u_rank],br,&tmp[child2->u_rows],br2);
          x[ch2]=new T[br2*d];
          gemm('C','N',br2,d,br2,ONE,child2->Q,br2,tmp,br2,ZERO,x[ch2],br2);
          delete[] tmp;
        } else {
          /* Q{child 2} = I (not stored), y{child 2} empty => x{child 2}=x{i}_2 */
          x[ch2]=new T[br2*d];
          lacpy('N',br2,d,&x[i][child1->u_rank],br,x[ch2],br2);
        }
      } else {
        /* Copy the piece of solution x[i] to the corresponding part
         * of the final solution X. */
        lacpy('N',br,d,x[i],br,&X[ib],a_dim);
      }

    }

    for(i=0;i<dim_tree-1;i++) {
      delete[] x[i];
      delete[] y[i];
    }
    delete[] x;
    delete[] y;
    y=NULL;
  }

}

#endif // HSS_SEQ_SOLVE_HPP
