/*
 * STRUMPACK -- STRUctured Matrices PACKage, Copyright (c) 2014, The Regents of
 * the University of California, through Lawrence Berkeley National Laboratory
 * (subject to receipt of any required approvals from the U.S. Dept. of Energy).
 * All rights reserved.
 *
 * If you have questions about your rights to use or distribute this software,
 * please contact Berkeley Lab's Technology Transfer Department at TTD@lbl.gov.
 *
 * NOTICE. This software is owned by the U.S. Department of Energy. As such, the
 * U.S. Government has been granted for itself and others acting on its behalf a
 * paid-up, nonexclusive, irrevocable, worldwide license in the Software to
 * reproduce, prepare derivative works, and perform publicly and display publicly.
 * Beginning five (5) years after the date permission to assert copyright is
 * obtained from the U.S. Department of Energy, and subject to any subsequent five
 * (5) year renewals, the U.S. Government is granted for itself and others acting
 * on its behalf a paid-up, nonexclusive, irrevocable, worldwide license in the
 * Software to reproduce, prepare derivative works, distribute copies to the
 * public, perform publicly and display publicly, and to permit others to do so.
 *
 * Developers: Francois-Henry Rouet, Xiaoye S. Li, Pieter Ghysels
 *             (Lawrence Berkeley National Lab, Computational Research Division).
 *
 */

#ifndef PAR_GENHMATRIX_HPP
#define PAR_GENHMATRIX_HPP

template<typename T, typename S> void par_genHMatrix(T *A, HSS_tree<T,S> *htree, S perrank, gridinfo *ginfo, int *descA) {
  /* Generate a kinda H-matrix where each block has rank perrank times its size.
   * The algorithm is the following:
   *   for each node i in postorder
   *     I=segment associated with i
   *     if i is a leaf
   *       A(I,I)=random matrix of size I + strong diagonal (full rank, and good for factorization)
   *     else
   *       b=length(I)
   *       I{1,2}=segment of child {1,2}
   *       A(I1,I2)= outer product of two random matrices of size b x (perrank*b) (should have rank roughly perrank*b)
   *
   * NB: this is for testing purposes, so it's not a method of HSS_tree.
   */

  int i, j;
  int ib, ie;
  int b, br, bc;
  int r;
  int dim_tree;
  int n;
  int ctxt, nprow, npcol, myrow, mycol, mb, nb;
  int descD[BLACSCTXTSIZE], descU[BLACSCTXTSIZE], descV[BLACSCTXTSIZE];
  int locr, locc;
  int idummy, ierr;
  T *U, *V, *D;
  HSS_node<T,S> **nodes;
  HSS_node<T,S> *child1, *child2;

  T ZERO=0;
  T  ONE=1;

  n=htree->a_dim;
  dim_tree=htree->dim_tree;
  nodes=htree->nodes;

  ctxt=ginfo->ctxt;
  nprow=ginfo->nprow;
  npcol=ginfo->npcol;
  myrow=ginfo->myrow;
  mycol=ginfo->mycol;
  mb=ginfo->mb;
  nb=ginfo->nb;

  StrumpackRand rgen(time(NULL)+myrow*npcol+mycol);

  for(i=0;i<dim_tree;i++) {
    ib=nodes[i]->tree_ib;
    ie=nodes[i]->tree_ie;
    b=ie-ib+1;
    if(nodes[i]->nr_children==0) {
      locr=numroc_(&b,&mb,&myrow,&IZERO,&nprow);
      locc=numroc_(&b,&nb,&mycol,&IZERO,&npcol);
      /* Diagonal blocks are full-rank with large values on the diagonal */
      D=new T[locr*locc];
      idummy=std::max(1,locr);
      descinit_(descD,&b,&b,&mb,&nb,&IZERO,&IZERO,&ctxt,&idummy,&ierr);
      parDenseRand(D,b,b,mb,nb,myrow,mycol,nprow,npcol,&rgen);
      for(j=1;j<=b;j++)
        if(myrow==indxg2p_(&j,&mb,&myrow,&IZERO,&nprow) && mycol==indxg2p_(&j,&nb,&mycol,&IZERO,&npcol))
        D[(indxg2l_(&j,&nb,&mycol,&IZERO,&npcol)-1)*locr+indxg2l_(&j,&mb,&myrow,&IZERO,&nprow)-1]=static_cast<T>(n)*static_cast<T>(n);
      pgeadd('N',b,b,ONE,D,IONE,IONE,descD,ZERO,A,ib,ib,descA);
      delete[] D;
    } else {
      /* Generate the two off-diagonal blocks A(I1,I2) and A(I2,I1) */
      child1=nodes[i]->children[0];
      child2=nodes[i]->children[1];

      /* A(I1,I2) */
      br=child1->tree_ie-child1->tree_ib+1;
      bc=child2->tree_ie-child2->tree_ib+1;
      /* Rank r: at least one, at most perrank*min(br,bc) */
      r=std::max(1,static_cast<int>(floor(perrank*(std::min(br,bc)))));
      /* U is br*r, V is r*bc */
      locr=numroc_(&br,&mb,&myrow,&IZERO,&nprow);
      locc=numroc_(&r,&nb,&mycol,&IZERO,&npcol);
      U=new T[locr*locc];
      idummy=std::max(1,locr);
      descinit_(descU,&br,&r,&mb,&nb,&IZERO,&IZERO,&ctxt,&idummy,&ierr);
      parDenseRand(U,br,r,mb,nb,myrow,mycol,nprow,npcol,&rgen);
      locr=numroc_(&r,&mb,&myrow,&IZERO,&nprow);
      locc=numroc_(&bc,&nb,&mycol,&IZERO,&npcol);
      V=new T[locr*locc];
      idummy=std::max(1,locr);
      descinit_(descV,&r,&bc,&mb,&nb,&IZERO,&IZERO,&ctxt,&idummy,&ierr);
      parDenseRand(V,r,bc,mb,nb,myrow,mycol,nprow,npcol,&rgen);
      pgemm<T>('N','N',br,bc,r,ONE,U,IONE,IONE,descU,V,IONE,IONE,descV,ZERO,A,child1->tree_ib,child2->tree_ib,descA);
      delete[] U;
      delete[] V;

      /* A(I2,I1) */
      br=child2->tree_ie-child2->tree_ib+1;
      bc=child1->tree_ie-child1->tree_ib+1;
      r=std::max(1,static_cast<int>(floor(perrank*(std::min(br,bc)))));
      /* U is br*r, V is r*bc */
      locr=numroc_(&br,&mb,&myrow,&IZERO,&nprow);
      locc=numroc_(&r,&nb,&mycol,&IZERO,&npcol);
      U=new T[locr*locc];
      idummy=std::max(1,locr);
      descinit_(descU,&br,&r,&mb,&nb,&IZERO,&IZERO,&ctxt,&idummy,&ierr);
      parDenseRand(U,br,r,mb,nb,myrow,mycol,nprow,npcol,&rgen);
      locr=numroc_(&r,&mb,&myrow,&IZERO,&nprow);
      locc=numroc_(&bc,&nb,&mycol,&IZERO,&npcol);
      V=new T[locr*locc];
      idummy=std::max(1,locr);
      descinit_(descV,&r,&bc,&mb,&nb,&IZERO,&IZERO,&ctxt,&idummy,&ierr);
      parDenseRand(V,r,bc,mb,nb,myrow,mycol,nprow,npcol,&rgen);
      pgemm<T>('N','N',br,bc,r,ONE,U,IONE,IONE,descU,V,IONE,IONE,descV,ZERO,A,child2->tree_ib,child1->tree_ib,descA);
      delete[] U;
      delete[] V;
    }
  }

}

#endif // PAR_GENHMATRIX_HPP
