/*
 * STRUMPACK -- STRUctured Matrices PACKage, Copyright (c) 2014, The Regents of
 * the University of California, through Lawrence Berkeley National Laboratory
 * (subject to receipt of any required approvals from the U.S. Dept. of Energy).
 * All rights reserved.
 *
 * If you have questions about your rights to use or distribute this software,
 * please contact Berkeley Lab's Technology Transfer Department at TTD@lbl.gov.
 *
 * NOTICE. This software is owned by the U.S. Department of Energy. As such, the
 * U.S. Government has been granted for itself and others acting on its behalf a
 * paid-up, nonexclusive, irrevocable, worldwide license in the Software to
 * reproduce, prepare derivative works, and perform publicly and display publicly.
 * Beginning five (5) years after the date permission to assert copyright is
 * obtained from the U.S. Department of Energy, and subject to any subsequent five
 * (5) year renewals, the U.S. Government is granted for itself and others acting
 * on its behalf a paid-up, nonexclusive, irrevocable, worldwide license in the
 * Software to reproduce, prepare derivative works, distribute copies to the
 * public, perform publicly and display publicly, and to permit others to do so.
 *
 * Developers: Pieter Ghysels, Francois-Henry Rouet, Xiaoye S. Li.
 *             (Lawrence Berkeley National Lab, Computational Research Division).
 */
/*! \file StrumpackSparseSolver.hpp
 * \brief Contains the definition of the sequential/multithreaded sparse solver class.
 */
#ifndef STRUMPACK_SOLVER_H
#define STRUMPACK_SOLVER_H

#include <sstream>
#include <getopt.h>
#include <new>
#include <cmath>
#ifdef USE_TBB_MALLOC
#include <tbb/scalable_allocator.h>
#endif
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#ifdef HAVE_PAPI
#include <papi.h>
#endif
#include "strumpack_parameters.hpp"
#include "CompressedSparseMatrix.hpp"
#include "CSRMatrix.hpp"
#include "MatrixReordering.hpp"
#include "MetisReordering.hpp"
#include "ScotchReordering.hpp"
#include "GeometricReordering.hpp"
#include "EliminationTree.hpp"
#include "tools.hpp"
#include "GMRes.hpp"
#include "BiCGStab.hpp"
#include "IterativeRefinement.hpp"

#ifdef USE_TBB_MALLOC
void* operator new(std::size_t sz) throw(std::bad_alloc) {
  return scalable_malloc(sz);
}
void operator delete(void* ptr) throw() {
  if (ptr==NULL) return;
  scalable_free(ptr);
}
#endif

/*! All of STRUMPACK-sparse is contained in the strumpack namespace. */
namespace strumpack {

  /*! \brief StrumpackSparseSolver<scalar_t,real_t,integer_t> is the
   *         main sequential or multithreaded sparse solver class.
   *
   * The StrumpackSparseSolver<scalar_t,real_t,integer_t> class
   * depends on 3 template paramaters. Supported combinations for
   * scalar_t,real_t are: <float,float>, <double,double>,
   * <std::complex<float>,float> and
   * <std::complex<double>,double>. The integer_t type defaults to a
   * regular int. If regular int causes 32 bit integer overflows, you
   * should switch to integer_t=int64_t instead.
   */
  template<typename scalar_t,typename real_t,typename integer_t=int> class StrumpackSparseSolver {
  public:
    /*! \brief Constructor of the StrumpackSparseSolver class,
     *         taking command line arguments.
     *
     * \param argc  The number of arguments, i.e, number of elements in the argv array.
     * \param argv  Command line arguments. Add -h or --help to have a description printed.
     * \param verb  Flag to enable output.
     * \param root  Flag to denote whether this process is the root MPI process.
     *              Only the root will print certain messages.
     */
    StrumpackSparseSolver(int argc, char* argv[], bool verb=true, bool root=true);
    /*! \brief Destructor of the StrumpackSparseSolver class. */
    virtual ~StrumpackSparseSolver();

    /*! \brief Parse the command line options passed to the
     *         constructor.
     *
     * The options are only parsed when this function is called, not
     * during construction of the class. This way, one can set options
     * on the object after construction, but before parsing the
     * command line options.
     */
    void set_from_options();
    /*! \brief Associate a (sequential) CSRMatrix with this solver.
     *
     * This matrix will not be modified. An internal copy will be
     * made, so it is safe to delete the data immediately after
     * calling this function. When this is called on a
     * StrumpackSparseSolverMPIDist object, the input matrix will
     * immediately be distributed over all the processes in the
     * communicator associated with the solver. This method is thus
     * collective on the MPI communicator associated with the solver.
     *
     * \param A  A CSRMatrix<scalar_t,real_t,integer_t> object.
     */
    virtual void set_matrix(CSRMatrix<scalar_t,real_t,integer_t>& A);
    /*! \brief Associate a (sequential) NxN CSR matrix with this
     *         solver.
     *
     * This matrix will not be modified. An internal copy will be
     * made, so it is safe to delete the data immediately after
     * calling this function. When this is called on a
     * StrumpackSparseSolverMPIDist object, the input matrix will
     * immediately be distributed over all the processes in the
     * communicator associated with the solver. This method is thus
     * collective on the MPI communicator associated with the solver.
     *
     * \param N                 The number of rows and columns of the CSR input matrix.
     * \param row_ptr           Indices in col_ind and values for the start of each row.
     *                          Nonzeros for row r are in [row_ptr[r],row_ptr[r+1]).
     * \param col_ind           Column indices of each nonzero.
     * \param values            Nonzero values.
     * \param symmetric_pattern Denotes whether the sparsity pattern of the input matrix is symmetric.
     */
    virtual void set_csr_matrix(integer_t N, integer_t* row_ptr, integer_t* col_ind, scalar_t* values, bool symmetric_pattern=false);
    /*! \brief Compute matrix reorderings for numerical stability and
     *         to reduce fill-in.
     *
     * Start computation of the matrix reorderings. See the relevant
     * options to control the matrix reordering. A first reordering is
     * the MC64 column permutation for numerical stability. This can
     * be disabled if the matrix has large nonzero diagonal
     * entries. MC64 optionally also performs row and column
     * scaling. Next, a fill-reducing reordering is computed. This is
     * done with the nested dissection algortihms of either
     * (PT-)Scotch, (Par)Metis or a simple geometric nested dissection
     * code which only works on regular meshes.
     *
     * \param nx  This (optional) parameter is only meaningful when the matrix corresponds
     *            to a stencil on a regular mesh. The stecil is assumed to be at most 3 points
     *            wide in each dimension and only contain a single degree of freedom per grid point.
     *            The nx parameter denotes the number of grid points in the first spatial dimension.
     * \param ny  See parameters nx. Parameter ny denotes the number of gridpoints in the second
     *            spatial dimension. This should only be set if the mesh is 2 or 3 dimensional.
     * \param nz  See parameters nx. Parameter nz denotes the number of gridpoints in the third
     *            spatial dimension. This should only be set if the mesh is 3 dimensional.
     * \return    Error code.
     * \sa        set_mc64job, set_matrix_reordering_method, set_nested_dissection_parameter,
     *            set_scotch_strategy
     */
    virtual params::returnCode reorder(int nx=1, int ny=1, int nz=1);
    /*! \brief Perform numerical factorization of the sparse input matrix.
     *
     * This is the computationally expensive part of the code.
     * \return    Error code.
     */
    params::returnCode factor();
    /*! \brief  Solve a linear system with a single right-hand side.
     *
     * \param b  Input, will not be modified. Pointer to the right-hand side.
     *           Array should be lenght N, the dimension of the input matrix for
     *           StrumpackSparseSolver and StrumpackSparseSolverMPI. For
     *           StrumpackSparseSolverMPIDist, the length of b should be correspond
     *           the partitioning of the block-row distributed input matrix.
     * \param x  Output. Pointer to the solution vector.
     *           Array should be lenght N, the dimension of the input matrix for
     *           StrumpackSparseSolver and StrumpackSparseSolverMPI. For
     *           StrumpackSparseSolverMPIDist, the length of b should be correspond
     *           the partitioning of the block-row distributed input matrix.
     * \param use_initial_guess Set to true if x contains an intial guess to the solution.
     *                          This is mainly useful when using an iterative solver.
     *                          If set to false, x should not be set (but should be allocated).
     * \return    Error code.
     */
    virtual params::returnCode solve(scalar_t* b, scalar_t* x, bool use_initial_guess=false);

    /*! \brief Set the maximum number of Krylov iterations.
     * \param maxit The maximum number of iterations, should be >= 1. */
    inline void set_maximum_Krylov_iterations(int maxit) { maximum_Krylov_iterations = maxit; }
    /*! \brief Set the GMRes restart length. A larger value most likely leads
     *         to faster convergence, but requires more memory.
     * \param m The new GMRes restart length, should be >= 1.  */
    inline void set_gmres_restart(int m) { gmres_restart = m; }
    /*! \brief Set the relative stopping tolerance for the outer Krylov solver (or iterative refinement).
    * \param tol Tolerance used when checking the relative decrease of the residual norm. */
    inline void set_relative_Krylov_tolerance(real_t tol) { relative_Krylov_tolerance = tol; }
    /*! \brief Set the absolute stopping tolerance for the outer Krylov solver (or iterative refinement).
    * \param tol Tolerance used when checking the absolute norm of the residual. */
    inline void set_absolute_Krylov_tolerance(real_t tol) { absolute_Krylov_tolerance = tol; }
    /*! \brief Set the Scotch strategy to use.
     * \param strat Currently only the values 1 or 2. */
    inline void set_scotch_strategy(int strat) { scotch_strategy = strat; }
    /*! \brief Set parameter controlling the leaf size in the nested dissection procedure.
    * \param nd_param  Stop the nested dissection recursion when a separator is smaller then nd_param. */
    inline void set_nested_dissection_parameter(int nd_param) { nested_dissection_parameter = nd_param; }
    /*! \brief Set the fill-reducing matrix reordering strategy
     * \param m   One of strumpack::params::MatrixReorderingStrategy. \n
     *            - METIS: Use Metis \n
     *            - PARMETIS: Use ParMetis \n
     *            - SCOTCH: Use Scotch \n
     *            - PTSCOTCH: Use PT-Scotch \n
     *            - RCM: Use reverse Cuthill-McKee \n
     *            - GEOMETRIC: Use a simple geometric nested dissection code for regular meshes.
     *              This only works when this->reorder(nx,ny,nz) was called with the correct mesh sizes.
     */
    inline void set_matrix_reordering_method(params::MatrixReorderingStrategy m) { reordering_method = m; }
    /*! \brief Set the type of Gram Schmidth orthogonalization used on GMRes.
     * \param t CLASSICAL is faster (more scalable) but MODIFIED is stable. */
    inline void set_GramSchmidt_type(params::GramSchmidtType t) { Gram_Schmidt_type = t; }
    /*! \brief Set the MC64 job type.
     *
     * MC64 is used to permute the input matrix for numerical
     * stability. It can also apply row and column scaling for
     * stability.
     *
     * \param job The job type. \n
     *        - job=0: Disable MC64, for many matrices MC64 is not required. \n
     *        - job=1: This is not supported. \n
     *        - job=2: Maximize the smallest diagonal value. \n
     *        - job=3: Same as 2, but using a different algorithm. \n
     *        - job=4: Maximize the sum of the diagonal entries. \n
     *        - job=5: Maximize the product of the diagonal entries and perform row and column scaling.
     */
    inline void set_mc64job(int job) { if (job < 0 || job > 5) std::cerr << "# WARNING: invalid mc64 job number" << std::endl; params::mc64job = job; }
    /*! \brief Set the outer iterative (Krylov) solver to be used.
     * \param solver_type No outer solver, iterative refinement, choice of Krylov solver, or auto. */
    inline void set_Krylov_solver(params::KrylovSolver solver_type) { Krylov_solver = solver_type; }
    /*! \brief Enable/disable HSS compression for large frontal matrices.
     * \sa set_HSS_minimum_front_size */
    inline void use_HSS(bool h) { hss_strategy.use_hss = h; }
    /*! \brief Set the minimum frontal matrix size to use HSS compression.
     * \sa use_HSS */
    inline void set_minimum_HSS_size(int s) { params::hss_minimum_front_size = s; }
    /*! \brief Set the relative compression tolerance used in rank-revealing QR factorization during HSS construction. */
    inline void set_relative_compression_tolerance(real_t rctol) { hss_strategy.rctol = rctol; }
    /*! \brief Set the absolute compression tolerance used in rank-revealing QR factorization during HSS construction. */
    inline void set_absolute_compression_tolerance(real_t actol) { hss_strategy.actol = actol; }
    /*! \brief Set the rank pattern. See section 5 of the manual. */
    inline void set_rank_pattern(params::RankPattern p) { hss_strategy.rank_pattern = p; }
    /*! \brief Disable/enable output of statistics. */
    inline void set_verbose(bool v=true) { verbose = v; }
    /*! \brief Get the maximum number of Krylov iterations. */
    inline int get_maximum_Krylov_iterations() { return maximum_Krylov_iterations; }
    /*! \brief Get the GMRes restart parameter. */
    inline int get_gmres_restart() { return gmres_restart; }
    /*! \brief Get the relative tolerance for the outer (Krylov) iterative solver */
    inline real_t get_relative_Krylov_tolerance() { return relative_Krylov_tolerance; }
    /*! \brief Get the absolute tolerance for the outer (Krylov) iterative solver */
    inline real_t get_absolute_Krylov_tolerance() { return absolute_Krylov_tolerance; }
    /*! \brief Get the strategy being used for Scotch/PT-Scotch. */
    inline int get_scotch_strategy() { return scotch_strategy; }
    /*! \brief Get the nested dissection parameter being used. (maximum size of nested dissection leaf separator), */
    inline int get_nested_dissection_parameter() { return nested_dissection_parameter; }
    /*! \brief Get the matrix reordering (nested dissection) method being used. */
    inline params::MatrixReorderingStrategy get_matrix_reordering_method() { return reordering_method; }
    /*! \brief Get the Gram-Schmidt type being used in GMRes. */
    inline params::GramSchmidtType get_GramSchmidt_type() { return Gram_Schmidt_type; }
    /*! \brief Get the maximum rank encountered in any of the HSS matrices.
     * Call this AFTER numerical factorization. */
    inline int get_maximum_rank() { return elimination_tree()->get_maximum_rank(); }
    /*! \brief Get the number of nonzeros in the (sparse) factors. This is the fill-in.
     * Call this AFTER numerical factorization. */
    inline long long get_factor_nonzeros() { return elimination_tree()->get_factor_nonzeros(); }
    /*! \brief Get the number of nonzeros in the (sparse) factors. This is the fill-in.
     * Call this AFTER numerical factorization. */
    inline long long get_factor_memory() { return elimination_tree()->get_factor_nonzeros() * sizeof(scalar_t); }
    /*! \brief Get the number of iterations performed by the outer (Krylov) iterative solver. */
    inline int get_Krylov_iterations() { return Krylov_iterations; }
    /*! \brief Get the job number being used for MC64. */
    inline int get_mc64job() { return params::mc64job; }
    /*! \brief Is HSS compression being used? */
    inline bool use_HSS() { return hss_strategy.use_hss; }
    /*! \brief Get the setting for the minimum size of a frontal matrix for which to use HSS compression. */
    inline int get_minimum_HSS_size() { return params::hss_minimum_front_size; }
    /*! \brief Get the relative compression tolerance used in rank-revealing QR during HSS construction. */
    inline real_t get_relative_compression_tolerance() { return hss_strategy.rctol; }
    /*! \brief Get the absolute compression tolerance used in rank-revealing QR during HSS construction. */
    inline real_t get_absolute_compression_tolerance() { return hss_strategy.actol; }
    /*! \brief Get the type of outer (Krylov) iterative solver being used. */
    inline params::KrylovSolver get_Krylov_solver() { return Krylov_solver; }
    /*! \brief Get the rank pattern. See section 5 of the manual. */
    inline params::RankPattern get_rank_pattern() { return hss_strategy.rank_pattern; }

  protected:
    virtual void setup_elimination_tree();
    virtual void setup_matrix_reordering();
    virtual int compute_reordering(int nx, int ny, int nz);
    virtual void compute_separator_reordering();
    virtual CompressedSparseMatrix<scalar_t,real_t,integer_t>* matrix() { return _mat; }
    virtual MatrixReordering<scalar_t,real_t,integer_t>* reordering() { return _nd; }
    virtual EliminationTree<scalar_t,real_t,integer_t>* elimination_tree() { return _et; }
    void papi_initialize();
    inline long long get_dense_factor_nonzeros() { return elimination_tree()->get_dense_factor_nonzeros(); }
    inline long long get_dense_factor_memory() { return elimination_tree()->get_dense_factor_nonzeros() * sizeof(scalar_t); }
    void print_solve_stats(TaskTimer& t);
    virtual void perf_counters_start();
    virtual void perf_counters_stop(std::string s);
    virtual void synchronize() {}
    virtual void communicate_ordering() {}

    int maximum_Krylov_iterations;
    int gmres_restart;
    real_t relative_Krylov_tolerance;
    real_t absolute_Krylov_tolerance;
    int scotch_strategy;
    int nested_dissection_parameter;
    HSSStrategy<real_t> hss_strategy;
    int Krylov_iterations;
    params::KrylovSolver Krylov_solver;
    params::GramSchmidtType Gram_Schmidt_type;
    params::MatrixReorderingStrategy reordering_method;

    std::vector<integer_t> mc64_cperm;
    std::vector<scalar_t> mc64_Dr; // row scaling
    std::vector<scalar_t> mc64_Dc; // column scaling

    bool verbose;
    bool is_root;
    std::new_handler old_handler;
    int argc_;
    char** argv_;
    std::ostream* rank_out;

    bool factored;
    bool reordered;

#if defined(HAVE_PAPI)
    float rtime0=0., ptime0=0.;
    long_long flpops0=0;
#endif
#if defined(COUNT_FLOPS)
    long long int f0, flops_total;
    long long int b0, bytes_total;
#endif
  private:
    CSRMatrix<scalar_t,real_t,integer_t>* _mat;
    MatrixReordering<scalar_t,real_t,integer_t>* _nd;
    EliminationTree<scalar_t,real_t,integer_t>* _et;
  };

  template<typename scalar_t,typename real_t,typename integer_t>
  StrumpackSparseSolver<scalar_t,real_t,integer_t>::StrumpackSparseSolver
  (int argc, char* argv[], bool verb, bool root) :
    maximum_Krylov_iterations(500),
    gmres_restart(30),
    relative_Krylov_tolerance(params::get_default_Krylov_relative_tolerance<real_t>()),
    absolute_Krylov_tolerance(params::get_default_Krylov_absolute_tolerance<real_t>()),
    scotch_strategy(2),
    nested_dissection_parameter(8),
    Krylov_iterations(0),
    Krylov_solver(params::AUTO),
    Gram_Schmidt_type(params::MODIFIED),
    reordering_method(params::METIS),
    verbose(verb), is_root(root),
    old_handler(std::set_new_handler([]{
	  std::cerr << "STRUMPACK: out of memory!" << std::endl; abort();
	})), argc_(argc), argv_(argv), rank_out(NULL),
    factored(false), reordered(false),
    _mat(nullptr), _nd(nullptr), _et(nullptr) {
    papi_initialize();
    if (verbose && is_root) {
      std::cout << "# Initializing STRUMPACK" << std::endl;
#if defined(_OPENMP)
      if (params::num_threads == 1)
	std::cout << "# using " << params::num_threads << " OpenMP thread" << std::endl;
      else std::cout << "# using " << params::num_threads << " OpenMP threads" << std::endl;
#else
      std::cout << "# running serially, no OpenMP support!" << std::endl;
#endif
      // a heuristic to set the recursion task cutoff level based on the number of threads
      if (params::num_threads == 1) params::task_recursion_cutoff_level = 0;
      else {
	params::task_recursion_cutoff_level = std::log2(params::num_threads) + 3;
	std::cout << "# number of tasking levels = " << params::task_recursion_cutoff_level
		  << " = log_2(#threads) + 3"<< std::endl;
      }
    }
#ifdef COUNT_FLOPS
    // TODO why does this not compile on my GCC 5.0.1??
    //if (!params::flops.is_lock_free()) std::cerr << "# WARNING: the flop counter is not lock free" << std::endl;
#endif
  }

  template<typename scalar_t,typename real_t,typename integer_t>
  StrumpackSparseSolver<scalar_t,real_t,integer_t>::~StrumpackSparseSolver() {
    std::set_new_handler(old_handler);
    delete _nd;
    delete _et;
    delete _mat;
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  StrumpackSparseSolver<scalar_t,real_t,integer_t>::setup_elimination_tree() {
    if (_et) delete _et;
    _et = new EliminationTree<scalar_t,real_t,integer_t>(matrix(), reordering()->sep_tree.get(), hss_strategy.use_hss);
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  StrumpackSparseSolver<scalar_t,real_t,integer_t>::setup_matrix_reordering() {
    if (_nd) delete _nd;
    _nd = new MatrixReordering<scalar_t,real_t,integer_t>(matrix()->size());
  }

  template<typename scalar_t,typename real_t,typename integer_t> int
  StrumpackSparseSolver<scalar_t,real_t,integer_t>::compute_reordering(int nx, int ny, int nz) {
    return _nd->nested_dissection(reordering_method, _mat, scotch_strategy,
				  nested_dissection_parameter, verbose, nx, ny, nz);
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  StrumpackSparseSolver<scalar_t,real_t,integer_t>::compute_separator_reordering() {
    _nd->separator_reordering(_mat, verbose && is_root);
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  StrumpackSparseSolver<scalar_t,real_t,integer_t>::papi_initialize() {
#if defined(HAVE_PAPI)
    float mflops = 0.;
    int retval = PAPI_flops(&rtime0, &ptime0, &flpops0, &mflops);
    if (retval!=PAPI_OK) {
      std::cerr << "# WARNING: problem starting the PAPI performance counters:" << std::endl;
      switch (retval) {
      case PAPI_EINVAL: std::cerr << "#   - the counters were already started by something other than: PAPI_flips() or PAPI_flops()." << std::endl; break;
      case PAPI_ENOEVNT: std::cerr << "#   - the floating point operations, floating point instructions or total cycles event does not exist." << std::endl; break;
      case PAPI_ENOMEM: std::cerr << "#   - insufficient memory to complete the operation." << std::endl; break;
      default: std::cerr << "#   - some other error: " << retval << std::endl;
      }
    }
#endif
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  StrumpackSparseSolver<scalar_t,real_t,integer_t>::perf_counters_start() {
#if defined(HAVE_PAPI)
    float mflops=0., _rtime0=0., _ptime0=0.;
    long_long _flpops0=0; // cannot use class variables in openmp clause
#pragma omp parallel reduction(+:_flpops0) reduction(max:_rtime0) reduction(max:_ptime0)
    PAPI_flops(&_rtime0, &_ptime0, &_flpops0, &mflops);
    flpops0 = _flpops0; rtime0 = _rtime0; ptime0 = _ptime0;
#endif
#ifdef COUNT_FLOPS
    f0 = params::flops;
    b0 = params::bytes;
#endif
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  StrumpackSparseSolver<scalar_t,real_t,integer_t>::perf_counters_stop(std::string s) {
#if defined(HAVE_PAPI)
    float mflops=0., rtime1=0., ptime1=0.;
    long_long flpops1=0;
#pragma omp parallel reduction(+:flpops1) reduction(max:rtime1) reduction(max:ptime1)
    PAPI_flops(&rtime1, &ptime1, &flpops1, &mflops);
    if (verbose && is_root)
      std::cout << "# " << s << " PAPI stats:" << std::endl
		<< "#   - total flops = " << double(flpops1-flpops0) << std::endl
		<< "#   - flop rate = " <<  double(flpops1-flpops0)/(rtime1-rtime0)/1e9 << " GFlops/s" << std::endl
		<< "#   - real time = " << rtime1-rtime0 << " sec" << std::endl
		<< "#   - processor time = " << ptime1-ptime0 << " sec" << std::endl;
#endif
#ifdef COUNT_FLOPS
    flops_total = params::flops - f0;
    bytes_total = params::bytes - b0;
#endif
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  StrumpackSparseSolver<scalar_t,real_t,integer_t>::print_solve_stats(TaskTimer& t) {
    double tel = t.elapsed();
    if (verbose && is_root) {
      std::cout << "# DIRECT/GMRES solve:" << std::endl
		<< "#   - abs_tol = " << absolute_Krylov_tolerance
		<< ", rel_tol = " << relative_Krylov_tolerance
		<< ", restart = " << gmres_restart
		<< ", maxit = " << maximum_Krylov_iterations << std::endl
		<< "#   - number of Krylov iterations = " << Krylov_iterations << std::endl
		<< "#   - solve time = " << tel << std::endl;
#ifdef COUNT_FLOPS
      std::cout << "#   - total flops = " << double(flops_total) << std::endl
		<< "#   - flop rate = " << flops_total / tel / 1e9 << " GFlop/s" << std::endl
		<< "#   - bytes moved = " << double(bytes_total) / 1e6 << " MB" << std::endl
		<< "#   - byte rate = " << bytes_total / tel / 1e9 << " GByte/s" << std::endl
		<< "#   - solve arithmetic intensity = " << double(flops_total) / bytes_total
		<< " flop/byte" << std::endl;
#endif
    }
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  StrumpackSparseSolver<scalar_t,real_t,integer_t>::set_from_options() {
    option long_options[] = {
      {"sp_gstype",                required_argument, 0, 1},
      {"sp_reordering_method",     required_argument, 0, 2},
      {"sp_random_distribution",   required_argument, 0, 3},
      {"sp_random_engine",         required_argument, 0, 4},
      {"sp_atol",                  required_argument, 0, 5},
      {"sp_rtol",                  required_argument, 0, 6},
      {"sp_rctol",                 required_argument, 0, 7},
      {"sp_actol",                 required_argument, 0, 22},
      {"sp_maxit",                 required_argument, 0, 8},
      {"sp_restart",               required_argument, 0, 9},
      {"sp_hss_front_size",        required_argument, 0, 10},
      {"sp_nd_param",              required_argument, 0, 11},
      {"sp_minpart",               required_argument, 0, 12},
      {"sp_task_level",            required_argument, 0, 13},
      {"sp_random_blocksize",      required_argument, 0, 14},
      {"sp_rank_offset",           required_argument, 0, 15},
      {"sp_max_rank",              required_argument, 0, 16},
      {"sp_mc64job",               required_argument, 0, 17},
      {"sp_print_ranks",           required_argument, 0, 18},
      {"sp_q_power",               required_argument, 0, 19},
      {"sp_Krylov_solver",         required_argument, 0, 20},
      {"sp_separator_ordering_level", required_argument, 0, 21},
      {"sp_hss",                   no_argument, 0, 23},
      {"sp_log_etree",             no_argument, 0, 24},
      {"sp_log_ranks",             no_argument, 0, 27},
      {"sp_rank_pattern",          required_argument, 0, 25},
      {"sp_rank_factor",           required_argument, 0, 26},
      {"sp_use_METIS_NodeNDP",     no_argument, 0, 28},
      {"sp_use_METIS_NodeND",      no_argument, 0, 29},
      {"sp_verbose",               no_argument, 0, 'v'},
      {"sp_quiet",                 no_argument, 0, 'q'},
      {"help",                     no_argument, 0, 'h'},
      {NULL, 0, NULL, 0}
    };
    int c, option_index = 0;
    bool unrecognized_options = false;
    opterr = 0;
    while ((c = getopt_long_only(argc_, argv_, "hvq", long_options, &option_index)) != -1) {
      switch (c) {
      case 1: { std::string s; std::istringstream iss(optarg); iss >> s;
	  if (s.compare("modified") == 0) Gram_Schmidt_type = params::MODIFIED;
	  else if (s.compare("classical") == 0) Gram_Schmidt_type = params::CLASSICAL;
	  else std::cerr << "# WARNING: Gram-Schmidt type not recognized, use 'modified' or classical" << std::endl;
      } break;
      case 2: { std::string s; std::istringstream iss(optarg); iss >> s;
	  if (s.compare("metis") == 0) reordering_method = params::METIS;
	  else if (s.compare("parmetis") == 0) reordering_method = params::PARMETIS;
	  else if (s.compare("scotch") == 0) reordering_method = params::SCOTCH;
	  else if (s.compare("ptscotch") == 0) reordering_method = params::PTSCOTCH;
	  else if (s.compare("geometric") == 0) reordering_method = params::GEOMETRIC;
	  else if (s.compare("rcm") == 0) reordering_method = params::RCM;
	  else std::cerr << "# WARNING: matrix reordering strategy not recognized, use 'metis', 'parmetis', 'scotch', 'ptscotch' or 'geometric'" << std::endl;
      } break;
      case 3: { std::string s; std::istringstream iss(optarg); iss >> s;
	  if (s.compare("normal") == 0) params::random_distribution = params::NORMAL;
	  else if (s.compare("uniform") == 0) params::random_distribution = params::UNIFORM;
	  else std::cerr << "# WARNING: random number distribution not recognized, use 'normal' or 'uniform'" << std::endl;
      } break;
      case 4: { std::string s; std::istringstream iss(optarg); iss >> s;
	  if (s.compare("linear") == 0) params::random_engine = params::LINEAR;
	  else if (s.compare("mersenne") == 0) params::random_engine = params::MERSENNE;
	  else std::cerr << "# WARNING: random number engine not recognized, use 'linear' or 'mersenne'" << std::endl;
      } break;
      case 20: { std::string s; std::istringstream iss(optarg); iss >> s;
	  if (s.compare("auto") == 0) Krylov_solver = params::AUTO;
	  else if (s.compare("direct") == 0) Krylov_solver = params::DIRECT;
	  else if (s.compare("refinement") == 0) Krylov_solver = params::REFINE;
	  else if (s.compare("pgmres") == 0) Krylov_solver = params::PREC_GMRES;
	  else if (s.compare("gmres") == 0) Krylov_solver = params::GMRES;
	  else if (s.compare("pbicgstab") == 0) Krylov_solver = params::PREC_BICGSTAB;
	  else if (s.compare("bicgstab") == 0) Krylov_solver = params::BICGSTAB;
	  else std::cerr << "# WARNING: Krylov solver not recognized, using default" << std::endl;
      } break;
      case 5: { std::istringstream iss(optarg); iss >> absolute_Krylov_tolerance; } break;
      case 6: { std::istringstream iss(optarg); iss >> relative_Krylov_tolerance; } break;
      case 7: { std::istringstream iss(optarg); iss >> hss_strategy.rctol; } break;
      case 22: { std::istringstream iss(optarg); iss >> hss_strategy.actol; } break;
      case 8: { std::istringstream iss(optarg); iss >> maximum_Krylov_iterations; } break;
      case 9: { std::istringstream iss(optarg); iss >> gmres_restart; } break;
      case 10: { std::istringstream iss(optarg); iss >> params::hss_minimum_front_size; } break;
      case 11: { std::istringstream iss(optarg); iss >> nested_dissection_parameter; } break;
      case 12: { std::istringstream iss(optarg); iss >> params::hss_minimum_partition_size; } break;
      case 13: { std::istringstream iss(optarg); iss >> params::task_recursion_cutoff_level;
	  std::cout << "# number of tasking levels = " << params::task_recursion_cutoff_level << std::endl;
      } break;
      case 14: { std::istringstream iss(optarg); iss >> params::random_vector_blocksize; } break;
      case 15: { std::istringstream iss(optarg); iss >> params::rank_offset; } break;
      case 16: { std::istringstream iss(optarg); iss >> params::maximum_rank; } break;
      case 17: { std::istringstream iss(optarg); iss >> params::mc64job; } break;
      case 18: { std::string s; std::istringstream iss(optarg); iss >> s;
	  rank_out = (std::ostream*) new std::ofstream(s); } break;
      case 19: { std::istringstream iss(optarg); iss >> params::q_power; } break;
      case 21: { std::istringstream iss(optarg); iss >> params::separator_ordering_level; } break;
      case 23: { hss_strategy.use_hss = true; } break;
      case 24: { params::log_etree = true; } break;
      case 27: { params::log_ranks = true; } break;
      case 28: { params::use_METIS_NodeNDP = true; } break;
      case 29: { params::use_METIS_NodeNDP = false; } break;
      case 25: { std::string s; std::istringstream iss(optarg); iss >> s;
	  if (s.compare("adaptive") == 0) hss_strategy.rank_pattern = params::ADAPTIVE;
	  else if (s.compare("constant") == 0) hss_strategy.rank_pattern = params::CONSTANT;
	  else if (s.compare("sqrtN") == 0) hss_strategy.rank_pattern = params::SQRTN;
	  else if (s.compare("sqrtNlogN") == 0) hss_strategy.rank_pattern = params::SQRTNLOGN;
	  else if (s.compare("bisectcut") == 0) hss_strategy.rank_pattern = params::BISECTIONCUT;
	  else std::cerr << "# WARNING: Rank pattern not recognized, using default" << std::endl;
      } break;
      case 26: { std::istringstream iss(optarg); iss >> params::rank_factor; } break;
      case 'h': {
	if (is_root)
	  std::cout << "# strumpack options:" << std::endl
		    << "#    --sp_gstype modified|classical" << std::endl
		    << "#           Gram-Schmidt type for GMRES" << std::endl
		    << "#    --sp_Krylov_solver auto|direct|refinement|pgmres|gmres|pbicgstab|bicgstab" << std::endl
		    << "#           default: auto (refinement when no HSS, pgmres (preconditioned) with HSS compression)" << std::endl
		    << "#    --sp_reordering_method metis|scotch|parmetis|ptscotch|rcm|geometric" << std::endl
		    << "#           Code for nested dissection." << std::endl
		    << "#           Geometric only works on regular meshes and you need to provide the sizes." << std::endl
		    << "#    --sp_random_distribution uniform|normal" << std::endl
		    << "#    --sp_random_engine linear|mersenne" << std::endl
		    << "#    --sp_atol real_t (default " << absolute_Krylov_tolerance << ")" << std::endl
		    << "#           Krylov absolute (preconditioned) residual stopping tolerance" << std::endl
		    << "#    --sp_rtol real_t (default " << relative_Krylov_tolerance << ")" << std::endl
		    << "#           Krylov relative (preconditioned) residual stopping tolerance" << std::endl
		    << "#    --sp_rctol real_t (default " << hss_strategy.rctol << ")" << std::endl
		    << "#           HSS relative compression tolerance" << std::endl
		    << "#    --sp_actol real_t (default " << hss_strategy.actol << ")" << std::endl
		    << "#           HSS absolute compression tolerance" << std::endl
		    << "#    --sp_maxit int (default " << maximum_Krylov_iterations << ")" << std::endl
		    << "#           maximum Krylov iterations" << std::endl
		    << "#    --sp_restart int (default " << gmres_restart << ")" << std::endl
		    << "#           gmres restart length" << std::endl
		    << "#    --sp_hss_front_size int (default " << params::hss_minimum_front_size << ")" << std::endl
		    << "#           minimum size of front for HSS compression" << std::endl
		    << "#    --sp_nd_param int (default " << nested_dissection_parameter << ")" << std::endl
		    << "#    --sp_minpart int (default " << params::hss_minimum_partition_size << ")" << std::endl
		    << "#    --sp_task_level int (default " << params::task_recursion_cutoff_level << ")" << std::endl
		    << "#    --sp_random_blocksize int (default " << params::random_vector_blocksize << ")" << std::endl
		    << "#    --sp_rank_offset int (default " << params::rank_offset << ")" << std::endl
		    << "#    --sp_max_rank int (default " << params::maximum_rank << ")" << std::endl
		    << "#    --sp_mc64job int [0-5] (default " << params::mc64job << ")" << std::endl
		    << "#    --sp_print_ranks file (default no)" << std::endl
		    << "#           only for sequential fronts" << std::endl
		    << "#    --sp_q_power int (default 0)" << std::endl
		    << "#    --sp_separator_ordering_level (default " << params::separator_ordering_level << ")" << std::endl
		    << "#    --sp_hss (default no)" << std::endl
		    << "#           use HSS compression of fronts" << std::endl
		    << "#    --sp_rank_pattern adaptive|constant|sqrtN|sqrtNlogN|bisectcut" << std::endl
		    << "#           default: " << hss_strategy.get_rank_pattern() << std::endl
		    << "#    --sp_rank_factor float (default " << params::rank_factor << ")" << std::endl
		    << "#    --sp_log_etree (default no)" << std::endl
		    << "#           print out the nested dissection tree" << std::endl
		    << "#    --sp_log_ranks (default no)" << std::endl
		    << "#           print out info about HSS front sizes, nr of random vectors and HSS ranks" << std::endl
		    << "#    --sp_use_METIS_NodeNDP (default " << std::boolalpha << params::use_METIS_NodeNDP << ")" << std::endl
		    << "#           use undocumented Metis routine NodeNDP instead of NodeND" << std::endl
		    << "#    --sp_use_METIS_NodeND (default " << std::boolalpha << (!params::use_METIS_NodeNDP) << ")" << std::endl
		    << "#           use Metis routine NodeNDP instead of the undocumented NodeNDP" << std::endl
		    << "#    --help or -h" << std::endl
		    << "#    --sp_verbose or -v" << std::endl
		    << "#    --sp_quiet or -q" << std::endl;
	synchronize();
      } break;
      case 'v': verbose = true; break;
      case 'q': verbose = false; break;
      case '?': unrecognized_options = true; break;
      default: break;
      }
    }
    if (unrecognized_options && is_root)
      std::cerr << "# WARNING STRUMPACK: There were unrecognized options." << std::endl;
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  StrumpackSparseSolver<scalar_t,real_t,integer_t>::set_matrix
  (CSRMatrix<scalar_t,real_t,integer_t>& A) {
    if (_mat) delete _mat;
    _mat = A.clone();
    factored = reordered = false;
  }

  template<typename scalar_t,typename real_t,typename integer_t> void
  StrumpackSparseSolver<scalar_t,real_t,integer_t>::set_csr_matrix
  (integer_t N, integer_t* row_ptr, integer_t* col_ind, scalar_t* values, bool symmetric_pattern) {
    if (_mat) delete _mat;
    _mat = new CSRMatrix<scalar_t,real_t,integer_t>(N, row_ptr, col_ind, values, symmetric_pattern);
    factored = reordered = false;
  }

  template<typename scalar_t,typename real_t,typename integer_t> params::returnCode
  StrumpackSparseSolver<scalar_t,real_t,integer_t>::reorder(int nx, int ny, int nz) {
    if (matrix() == NULL) return params::MATRIX_NOT_SET;
    TaskTimer t1("permute-scale");
    int ierr;
    if (params::mc64job != 0) {
      if (verbose && is_root) {
	std::cout << "# mc64ad job " << params::mc64job << ": ";
	switch (params::mc64job) {
	case 1: std::cout << "maximum cardinality ! Doesn't work" << std::endl; return params::REORDERING_ERROR; break;
	case 2: std::cout << "maximum smallest diagonal value, version 1" << std::endl; break;
	case 3: std::cout << "maximum smallest diagonal value, version 2" << std::endl; break;
	case 4: std::cout << "maximum sum of diagonal values" << std::endl; break;
	case 5: std::cout << "maximum matching + row and column scaling" << std::endl; break;
	}
      }
      t1.time([&](){ ierr = matrix()->permute_and_scale(params::mc64job, mc64_cperm, mc64_Dr, mc64_Dc); });
      if (ierr) {
	std::cerr << "ERROR: mc64 failed" << std::endl;
	return params::REORDERING_ERROR;
      }
    }
    auto old_nnz = matrix()->nnz();
    TaskTimer t2("sparsity-symmetrization", [&](){ matrix()->symmetrize_sparsity(); });
    if (matrix()->nnz() != old_nnz && verbose && is_root)
      std::cout << "# Matrix padded with zeros to get symmetric pattern." << std::endl
		<< "# Number of nonzeros increased from "
		<< number_format_with_commas(old_nnz) << " to "
		<< number_format_with_commas(matrix()->nnz()) << "." << std::endl;

    TaskTimer t3("nested-dissection");
    perf_counters_start();
    t3.start();
    setup_matrix_reordering();
    ierr = compute_reordering(nx, ny, nz);
    if (ierr) { std::cerr << "ERROR: nested dissection went wrong, ierr=" << ierr << std::endl; return params::REORDERING_ERROR; }
    matrix()->permute(reordering()->iperm, reordering()->perm);
    t3.stop();
    if (verbose && is_root) {
      std::cout << "#   - nd time = " << t3.elapsed() << std::endl;
      if (params::mc64job != 0) std::cout << "#   - mc64 time = " << t1.elapsed() << std::endl;
      std::cout << "#   - symmetrization time = " << t2.elapsed() << std::endl;
    }
    perf_counters_stop("nested dissection");

    if (hss_strategy.use_hss) {
      perf_counters_start();
      TaskTimer t4("separator-reordering", [&](){
	  // reordering()->separator_reordering(matrix(), verbose && is_root);
	  compute_separator_reordering();
	  // TODO also broadcast this?? is computed with scotch
	  matrix()->permute(reordering()->sep_iorder, reordering()->sep_order);
	});
      if (verbose && is_root) std::cout << "#   - sep-reorder time = " << t4.elapsed() << std::endl;
      perf_counters_stop("separator reordering");
    }

    perf_counters_start();
    TaskTimer t0("symbolic-factorization", [&](){ setup_elimination_tree(); });
    reordering()->clear_tree_data();
    if (verbose) {
      auto nr_dense = elimination_tree()->get_nr_dense_fronts(); // this might require a reduction
      auto nr_HSS = elimination_tree()->get_nr_HSS_fronts();
      if (is_root)
	std::cout << "# symbolic factorization:" << std::endl
		  << "#   - nr of dense Frontal matrices = " << number_format_with_commas(nr_dense) << std::endl
		  << "#   - nr of HSS Frontal matrices = " << number_format_with_commas(nr_HSS) << std::endl
		  << "#   - symb-factor time = " << t0.elapsed() << std::endl;
    }
    perf_counters_stop("symbolic factorization");
    reordered = true;
    return params::SUCCESS;
  }

  template<typename scalar_t,typename real_t,typename integer_t> params::returnCode
  StrumpackSparseSolver<scalar_t,real_t,integer_t>::factor() {
    if (matrix() == NULL) return params::MATRIX_NOT_SET;
    if (!reordered) { params::returnCode ierr = reorder(); if (ierr) return ierr; }

    perf_counters_start();
    TaskTimer t1("factorization", [&]() { elimination_tree()->random_multifrontal_factorization(hss_strategy); });
    perf_counters_stop("numerical factorization");
    if (verbose) {
      auto fnnz = get_factor_nonzeros();
      auto dfnnz = get_dense_factor_memory();
      auto max_rank = get_maximum_rank();
      if (is_root) {
	std::cout << "# multifrontal factorization:" << std::endl
		  << "#   - factor time = " << t1.elapsed() << std::endl
		  << "#   - factor nonzeros = " << number_format_with_commas(fnnz) << std::endl
		  << "#   - factor memory = " << fnnz * sizeof(scalar_t) / 1e6 << " MB" << std::endl;
#ifdef COUNT_FLOPS
	std::cout << "#   - total flops = " << double(flops_total) << std::endl
		  << "#   - flop rate = " << flops_total / t1.elapsed() / 1e9 << " GFlop/s" << std::endl;
#endif
	std::cout << "#   - factor memory/nonzeros = " << float(fnnz * sizeof(scalar_t)) / dfnnz * 100.0 << " % of multifrontal" << std::endl
		  << "#   - maximum HSS rank = " << max_rank << std::endl
		  << "#   - HSS compression = " << std::boolalpha << hss_strategy.use_hss << std::endl
		  << "#   - HSS rank pattern = " << hss_strategy.get_rank_pattern() << std::endl
		  << "#   - relative compression tolerance = " << hss_strategy.rctol << std::endl
		  << "#   - absolute compression tolerance = " << hss_strategy.actol << std::endl;
	if (params::random_engine == params::LINEAR) {
	  if (params::random_distribution == params::NORMAL)
	    std::cout << "#   - normal(0,1) distribution with minstd_rand engine" << std::endl;
	  else if (params::random_distribution == params::UNIFORM)
	    std::cout << "#   - uniform [0,1] distribution with minstd_rand engine" << std::endl;
	} else if (params::random_engine == params::MERSENNE) {
	  if (params::random_distribution == params::NORMAL)
	    std::cout << "#   - normal(0,1) distribution with mt19937 engine" << std::endl;
	  else if (params::random_distribution == params::UNIFORM)
	    std::cout << "#   - uniform [0,1] distribution with mt19937 engine" << std::endl;
	}
      }
    }
    if (rank_out != NULL) elimination_tree()->print_rank_statistics(*rank_out);
    factored = true;
    return params::SUCCESS;
  }

  template<typename scalar_t,typename real_t,typename integer_t> params::returnCode
  StrumpackSparseSolver<scalar_t,real_t,integer_t>::solve
  (scalar_t* b, scalar_t* x, bool use_initial_guess) {
    if (!factored) { params::returnCode ierr = factor(); if (ierr) return ierr; }
    TaskTimer t("solve");
    perf_counters_start();
    t.start();
    auto N = matrix()->size();

    auto b_loc = new scalar_t[N];
    if (params::mc64job == 5) x_mult_y(N, b, mc64_Dr.data(), b_loc);
    else { std::copy(b, b+N, b_loc); STRUMPACK_BYTES(sizeof(scalar_t)*static_cast<long long int>(N)*3); }
    permute_vector(N, b_loc, reordering()->iperm, 1);
    if (hss_strategy.use_hss) permute_vector(N, b_loc, reordering()->sep_iorder, 1);

    if (!use_initial_guess) {
      std::fill(x, x+N, scalar_t(0.)); STRUMPACK_BYTES(sizeof(scalar_t)*static_cast<long long int>(N)*2);
    } else {
      if (params::mc64job == 5) x_div_y(N, x, mc64_Dc.data());
      if (params::mc64job != 0) permute_vector(N, x, mc64_cperm, 1);
      permute_vector(N, x, reordering()->iperm, 1);
      if (hss_strategy.use_hss) permute_vector(N, x, reordering()->sep_iorder, 1);
    }
    Krylov_iterations = 0;
    elimination_tree()->allocate_solve_work_memory(); // for 1 vector

    auto gmres_solve = [&](std::function<void(scalar_t*)> preconditioner) {
      GMRes<scalar_t,real_t,integer_t>
      (matrix(), preconditioner, N, x, b_loc, relative_Krylov_tolerance, absolute_Krylov_tolerance,
       Krylov_iterations, maximum_Krylov_iterations,
       gmres_restart, Gram_Schmidt_type, use_initial_guess, verbose && is_root);
    };
    auto bicgstab_solve = [&](std::function<void(scalar_t*)> preconditioner) {
      BiCGStab<scalar_t,real_t,integer_t>
      (matrix(), preconditioner, N, x, b_loc, relative_Krylov_tolerance, absolute_Krylov_tolerance,
       Krylov_iterations, maximum_Krylov_iterations, use_initial_guess, verbose && is_root);
    };
    auto loc_et = elimination_tree();
    auto refine = [&]() {
      IterativeRefinement<scalar_t,real_t,integer_t>
      (matrix(), [loc_et](scalar_t* x){ loc_et->multifrontal_solve(x); },
       N, x, b_loc, relative_Krylov_tolerance, absolute_Krylov_tolerance,
       Krylov_iterations, maximum_Krylov_iterations, use_initial_guess, verbose && is_root);
    };
    switch (Krylov_solver) {
    case params::AUTO: {
      if (hss_strategy.use_hss) gmres_solve([loc_et,N](scalar_t* x){ loc_et->multifrontal_solve(x); });
      else refine();
    }; break;
    case params::DIRECT: {
      std::copy(b_loc, b_loc+N, x);
      elimination_tree()->multifrontal_solve(x);
      STRUMPACK_BYTES(sizeof(scalar_t)*static_cast<long long int>(N)*3);
    }; break;
    case params::REFINE:     { refine(); }; break;
    case params::PREC_GMRES: { gmres_solve([loc_et](scalar_t* x){ loc_et->multifrontal_solve(x); }); }; break;
    case params::GMRES:      { gmres_solve([](scalar_t* x){}); }; break;
    case params::PREC_BICGSTAB: { bicgstab_solve([loc_et](scalar_t* x){ loc_et->multifrontal_solve(x); }); }; break;
    case params::BICGSTAB:      { bicgstab_solve([](scalar_t* x){}); }; break;
    }

    delete[] b_loc;
    elimination_tree()->delete_solve_work_memory();

    if (hss_strategy.use_hss) permute_vector(N, x, reordering()->sep_order, 1);
    permute_vector(N, x, reordering()->perm, 1);
    if (params::mc64job != 0) permute_vector(N, x, mc64_cperm, 0);
    if (params::mc64job == 5) x_mult_y(N, x, mc64_Dc.data());

    t.stop();
    perf_counters_stop("DIRECT/GMRES solve");
    print_solve_stats(t);
    return params::SUCCESS;
  }

} //end namespace strumpack

#endif
